#ifndef GBASYS_H
#define GBASYS_H

#define IN_EWRAM __attribute__ ((section (".ewram")))
#define IN_IWRAM __attribute__ ((section (".iwram")))
#define LONG_CALL __attribute__ ((long_call))

#include "GBAReg.h"

#define NULL ((void*)0)

//Palette
#define BG_PALETTE        ((unsigned short*)0x05000000)
#define OBJ_PALETTE       ((unsigned short*)0x05000200)

//VRAM
#define MAP_0             ((unsigned short*)0x06000000)
#define MAP_1             ((unsigned short*)0x06000800)
#define MAP_2             ((unsigned short*)0x06001000)
#define MAP_3             ((unsigned short*)0x06001800)
#define MAP_4             ((unsigned short*)0x06002000)
#define SCREEN_0K         ((unsigned char*) 0x06000000)
#define SCREEN_16K        ((unsigned short*)0x06004000)
#define SCREEN_32K        ((unsigned short*)0x06008000)
#define SCREEN_48K        ((unsigned short*)0x0600C000)

#define M4_FRAME0             ((unsigned short*) 0x06000000)
#define M4_FRAME1             ((unsigned short*) 0x0600A000)
#define SPRITE_DATA           ((unsigned short*)0x06010000)
#define SPRITE_DATA_BITMAPPED ((unsigned short*)0x06014000)

//OAM
typedef struct _Sprite {
  unsigned short Attrib0;
  unsigned short Attrib1;
  unsigned short Attrib2;
  unsigned short RotateScale;
} Sprite;

#define OAM                   ((Sprite*)0x07000000)

#define MODE_0	0x0
#define MODE_1	0x1
#define MODE_2	0x2
#define MODE_3	0x3
#define MODE_4	0x4
#define MODE_5	0x5

//define the buffer which is used to set the active buffer
//when using double buffering
#define backbuffer 	0x10

//This bit, when set allows OAM(Object Attribute Memory) to
//be updated during a horizontal blank
#define H_BLANK_OAM 	0x20

//use these two defines to choose which mapping mode is used
//for sprite graphics 2D or 1D
#define OBJ_MAP_2D 	0x0
#define OBJ_MAP_1D 	0x40

//Causes the screen to go white by using a forced blank
#define FORCE_BLANK 	0x80

//define the flags for enabling backgrounds and objects(sprites)
#define BG0_ENABLE	0x100
#define BG1_ENABLE	0x200
#define BG2_ENABLE	0x400
#define BG3_ENABLE	0x800
#define OBJ_ENABLE	0x1000

//allows window displays (dont worry about these)
#define WIN1_ENABLE	0x2000
#define WIN2_ENABLE	0x4000
#define WINOBJ_ENABLE	0x8000


//Set the mode that you want to use, logical AND them together as below:
//e.g. SetMode(MODE_2 | OBJ_ENABLE | OBJ_MAP_1D);
#define SetMode(mode) REG_DISPCNT = (mode)

//SRAM / EEPROM
//#define FLASH_1M              ((unsigned char*)0x09fe0000)          // 1M Flash Memory
//#define FLASH_1M_END          ((unsigned char*)0x0a000000)

#endif //GBASYS_H
