#define MULTIBOOT const int __gba_multiboot;
MULTIBOOT
#define ubyte unsigned char

#include "GBASys.h"
#include "GBAReg.h"


extern const int * Soundbuffer;			//	defined in "sidplay.s"
extern const int * Memory;				//	"
void C64Init(const int , const int );	//	"
void Fill_Sound_Buffer(const int);		//	"

//	playsid-file
#include "cybernoid_II.sid.h"				//	<---------------------------------  place psid-file here!
// define global variables we'll extract out of the .psid-header
unsigned short	load 	=	0;	// becomes c64-load-address
unsigned short	init 	=	0;	// init-address
unsigned short	play	=	0;	// play-address
unsigned short	songs 	=	0;	// no. of songs in file
unsigned short	start 	=	0;	// no. of song to start with
unsigned long	speed 	=	0;	// speed (currently not implemented)

//	definition for playsid-header
typedef struct
{
	//
	// All values in big-endian order.
	//
	char id[4];          // 'PSID'
	ubyte version[2];    // 0x0001 or 0x0002
	ubyte data[2];       // 16-bit offset to binary data in file
	ubyte load[2];       // 16-bit C64 address to load file to (remx: if null -> startaddress found at the beginning of the sid)
	ubyte init[2];       // 16-bit C64 address of init subroutine
	ubyte play[2];       // 16-bit C64 address of play subroutine
	ubyte songs[2];      // number of songs
	ubyte start[2];      // start song (1-256 !)
	ubyte speed[4];      // 32-bit speed info
		                 // bit: 0=50 Hz, 1=CIA 1 Timer A (default: 60 Hz)
	char name[32];       // ASCII strings, 31 characters long and
	char author[32];     // terminated by a trailing zero
	char copyright[32];  //
	ubyte flags[2];      // only version 0x0002
	ubyte relocStartPage;
	ubyte relocPages;
	ubyte reserved[4];   // only version 0x0002
} psidHeader;

// ---------------------------------------------------------------------------------------------

extern const unsigned short denosis_Tiles[32768];
extern const unsigned short master_Palette[256];
extern const short 			sinus[256];

// ---------------------------------------------------------------------------------------------

typedef volatile unsigned char  vu8;
typedef volatile unsigned short vu16;
typedef volatile unsigned long  vu32;
typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned long  u32;


#define REG_SOUNDCNT   *(vu32*)0x4000080
#define REG_SOUNDCNT_L *(vu16*)0x4000080		//DMG sound control
#define REG_SOUNDCNT_H *(vu16*)0x4000082		//Direct sound control
#define REG_SOUNDCNT_X *(vu16*)0x4000084	    //Extended sound control
#define REG_DMA1SAD     *(u32*)0x40000BC	//DMA1 Source Address
#define REG_DMA1DAD     *(u32*)0x40000C0	//DMA1 Desination Address
#define REG_DMA1CNT     *(u32*)0x40000C4	//DMA1 Control
#define REG_DMA1CNT_L   *(u16*)0x40000C4	//DMA1 Control Low Value
#define REG_DMA1CNT_H   *(u16*)0x40000C6	//DMA1 Control High Value
#define REG_TM0CNT_L	*(u16*)0x4000100	//Timer 0 count value
#define REG_TM0CNT_H    *(u16*)0x4000102	//Timer 0 Control
#define REG_TM1CNT_L    *(u16*)0x4000104		//Timer 2 count value
#define REG_TM1CNT_H    *(u16*)0x4000106		//Timer 2


// Formula for playback frequency is: 0xFFFF-round(cpuFreq/playbackFreq)
// cpuFreq is 16.78Mhz ~ 16780000hz / 16khz (16000) = ~0x0400
// 0xffff-0x0400 = 0xfc00
#define	PlaybackFrequency	0xfc00

//	--------------------------------------------------------------------------------------------
//	--------------------------------------------------------------------------------------------
//	--------------------------------------------------------------------------------------------
void DMAInit (void){
	//Play a mono sound at 16khz in DMA mode Direct Sound
	//uses timer 0 as sampling rate source
	//uses timer 1 to count the samples played in order to stop the sound
	REG_SOUNDCNT_X=0x0000;  //turn sound chip off (not really needed!?)
	REG_SOUNDCNT_L=0;
	REG_SOUNDCNT_H=0x0b0F;  //DS A&B + fifo reset + timer0 + max volume to L and R
	REG_SOUNDCNT_X=0x0080;  //turn sound chip on

	REG_DMA1SAD=(unsigned long)&Soundbuffer; 	//dma1 source
	REG_DMA1DAD=0x040000a0; 					//write to FIFO A address
	REG_IE=REG_IE|0x0200;	  					//enable dma1-irq
	REG_DMA1CNT_H=0xf600;						//dma control: DMA enabled+ start on FIFO+32bit+repeat


	// 50hz-playback:
	// 0xffff-(16780000/(50*1024)) = 0xfeb8
	REG_TM1CNT_L=0xfeb8;			//0xffff-the number of samples to play
	REG_TM1CNT_H=0xC3;				//enable timer1+overflow+1024clock
	REG_IE=REG_IE|0x10;	  				//enable irq for timer 1 overflow

	REG_TM0CNT_L=PlaybackFrequency;	//playback freq
	REG_TM0CNT_H=0x0080; 			//enable timer at CPU freq
}

//	--------------------------------------------------------------------------------------------
//	get load/init/play-address from psid-file and copy song to specified position in c64-memory
//	--------------------------------------------------------------------------------------------
void	CopySID() {
	// psid-header
	psidHeader *	psid =	(psidHeader *)		&binarysid;
	// pointer to psid
	unsigned char * data =  (unsigned char *)	((long)&binarysid)+((psid->data[0]<<8))+psid->data[1];
	// load-address in c64-memory:
	load =	(psid->load[0]<<8)+psid->load[1];
	// if load-address IS NULL, load address is before data begins:
	if(!load) {
		load	=	(data[1]<<8)+data[0];	// reverse order!
		data	+=2;
	}
	// get init:
	init =	(psid->init[0]<<8)+psid->init[1];
	// get play:
	play =	(psid->play[0]<<8)+psid->play[1];
	// get songs:
	songs =	(psid->songs[0]<<8)+psid->songs[1];
	// get start:
	start =	(psid->start[0]<<8)+psid->start[1];
	// get speed:
	speed =	(psid->speed[0]<<24)+(psid->speed[1]<<16)+(psid->speed[2]<<8)+(psid->speed[0]);

	//	copy song to c64-memory:
	unsigned		counter= 0;
	unsigned char*	buffer = (unsigned char*)((unsigned long)&Memory);

	for(counter=0;counter<sizeof(binarysid)-sizeof(psid);counter++) {
		buffer[counter+load] = data[counter];
	}
}


//	--------------------------------------------------------------------------------------------
//	Frequency-Table (used to find frequence of played note to display it on keyboard)
//	--------------------------------------------------------------------------------------------
const unsigned short frequencies[6*16] = {
	0x0116 - 4,		0x0127 - 4,		0x0139 - 4,		0x014B - 4,		0x015F - 5,		0x0174 - 5,
	0x018A - 6,		0x01A1 - 6,		0x01BA - 6,		0x01D4 - 7,		0x01F0 - 7,		0x020E - 8,
	0x022D - 8,		0x024E - 8,		0x0271 - 9,		0x0296 - 9,		0x02BE - 10,	0x02E7 - 11,
	0x0314 - 11,	0x0342 - 12,	0x0374 - 12,	0x03A9 - 13,	0x03E0 - 14,	0x041B - 15,
	0x045A - 16,	0x049C - 17,	0x04E2 - 18,	0x052D - 19,	0x057B - 20,	0x05CF - 21,
	0x0627 - 23,	0x0685 - 24,	0x06E8 - 26,	0x0751 - 27,	0x07C1 - 29,	0x0837 - 31,
	0x08B4 - 33,	0x0938 - 34,	0x09C4 - 36,	0x0A59 - 38,	0x0AF4 - 40,	0x0B9E - 43,
	0x0C4E - 46,	0x0D0A - 49,	0x0DD0 - 52,	0x0EA2 - 55,	0x0F81 - 58,	0x106D - 62,
	0x1167 - 65,	0x1270 - 70,	0x1389 - 73,	0x14B2 - 76,	0x15ED - 80,	0x173B - 83,
	0x199D - 87,	0x1A14 - 92,	0x1BA0 - 103,	0x1D45 - 110,	0x1F03 - 116,	0x20DB - 123,
	0x22CF - 130,	0x24E1 - 138,	0x2712 - 146,	0x2965 - 155,	0x2BDB - 164,	0x2E76 - 172,
	0x313A - 185,	0x3427 - 196,	0x3738 - 207,	0x3A8A - 220,	0x3E05 - 233,	0x41B5 - 246,
	0x459D - 261,	0x49C1 - 277,	0x4E24 - 293,	0x52C9 - 311,	0x57B6 - 330,	0x5CED - 350,
	0x6273 - 370,	0x684E - 392,	0x6E82 - 415,	0x7514 - 440,	0x7C0A - 466,	0x836A - 494,
	0x8B3B - 523,	0x9382 - 554,	0x9C48 - 587,	0xA593 - 644,	0xAF6B - 659,	0xB9DA - 700,
	0xC4E7 - 740,	0xD09C - 784,	0xDD04 - 830,	0xEA28 - 880,	0xF814 - 932,	0xFFFF
};

//	--------------------------------------------------------------------------------------------
//	draws "a line" of a keyboard (8 octaves)
//	--------------------------------------------------------------------------------------------
void vxKeyboardHelper(unsigned note, unsigned yoffset, unsigned voice) {
	unsigned x;
	unsigned o = (note/12);			//	octave
	unsigned y = note-(o*12);		//	note on octave
	unsigned gfx;

	Sprite	mySprite;
	for(x=0;x<8;x++) {
		if(x!=o) {
			// not our octave -> display "default"-one!
			gfx = 0;
		} else {
			// octave ok -> display right note
			gfx = ((y&3)*8)+(((y>>2)&3)*32);//+(((y>>4)&3)*8);
			if(y>=8) gfx+=64;	// must jump over a line...
		}

		mySprite.Attrib0 = 0x6000+(yoffset);
		mySprite.Attrib1 = 0x4000+(x*28)+8;
		mySprite.Attrib2 = gfx;
		mySprite.RotateScale = 0;
		OAM[x+(voice*16)] = mySprite;
		mySprite.Attrib0 = 0x6000+(yoffset+8);
		mySprite.Attrib1 = 0x4000+(x*28)+8;
		mySprite.Attrib2 = gfx+64;
		mySprite.RotateScale = 0;
		OAM[x+(voice*16)+8] = mySprite;
	}
	return;
}

//	--------------------------------------------------------------------------------------------
//	find currently played note
//	--------------------------------------------------------------------------------------------
void DrawKeyboard(int draw) {
	unsigned y,yoffset;
	unsigned short Frequency;
	unsigned char*	pSID = (unsigned char*)(0xd400+(unsigned long)&Memory);

	if(!draw) yoffset=0;
	else	 yoffset=80;		// draw keyboard offscreen

	//  find frequency of voice1:
	Frequency = (pSID[0]+(pSID[1]<<8));
	for(y=0;y<(6*16);y++)
		if (frequencies[y] >= Frequency) break;
	//  y should now be the note currently played!
	//	set played note:
	vxKeyboardHelper(y,80+yoffset,0);

	//  find frequency of voice2:
	Frequency = (pSID[7]+(pSID[8]<<8));
	for(y=0;y<(6*16);y++)
		if (frequencies[y] >= Frequency) break;
	//  y should now be the note currently played!
	vxKeyboardHelper(y,100+yoffset,1);

	//  find frequency of voice3:
	Frequency = (pSID[14]+(pSID[15]<<8));
	for(y=0;y<(6*16);y++)
		if (frequencies[y] >= Frequency) break;
	//  y should now be the note currently played!
	vxKeyboardHelper(y,120+yoffset,2);
}

char	szText[12][30] = {
	"------------------------------",
	"         HIHO FOLKS!          ",
	"  YOU'RE CURRENTLY LISTENING  ",
	"     TO CYBERNOID II FROM     ",
	"  JEROEN TEL/MANIACS OF NOISE ",
	"                              ",
	"    THIS SIDPLAYER WAS DONE   ",
	"     BY DENYS BOGATZ 2002.    ",
	"                              ",
	"     HOLD 'A' FOR KEYBOARD    ",
	"           ENJOY!!!           ",
	"------------------------------",
};
void	vxDisplayText() {
	unsigned	x,y;
    for(x=0;x<30;x++)
    for(y=0;y<12; y++) {
      MAP_0[x+((y+8)*32)] = ((szText[y][x]) & 0x3f) + (16*32);
  }

}

int	VOffset = 0x00;
int	HOffset = 0x00;
//	--------------------------------------------------------------------------------------------
//	background movement done in vblanc-irq
//	--------------------------------------------------------------------------------------------
void VBlanc(void){

	DrawKeyboard(REG_P1&1);

	// make it move
	REG_BG3VOFS = (sinus[VOffset & 0xff]) >> 4;
	REG_BG2VOFS = (sinus[VOffset & 0xff]) >> 3;
	REG_BG1VOFS = (sinus[VOffset & 0xff]) >> 2;

	REG_BG3HOFS = (sinus[HOffset & 0xff]) >> 4;
	REG_BG2HOFS = (sinus[HOffset & 0xff]) >> 3;
	REG_BG1HOFS = (sinus[HOffset & 0xff]) >> 2;

	VOffset+=2;
	HOffset+=3;

}

int iSamples 		= 0;
int iAktiveBuffer 	= 0;
int iFillBuffer	    = -1;
//	--------------------------------------------------------------------------------------------
//	callback-function for soundbuffer swaps!
//	--------------------------------------------------------------------------------------------
void DMA1IRQ(void) {
	iSamples++;
	if(iSamples>=0x10)	// 0x10 ~ 256 bytes (10*0x10) (DMA fills 32 Bytes per Request!)
	{
		iAktiveBuffer++;
		iFillBuffer	   = (iAktiveBuffer+1) & 0x0f;
		iAktiveBuffer &= 0x0f;

		// reset dma on max. buffers reached
		if(iAktiveBuffer==0) {
			// stop dma
			REG_DMA1CNT		=	0x0004c440;
			// have to wait 3cycles before writing to dma again...
			REG_SOUNDCNT_H  =	0x0b0f;	//DS A&B + fifo reset + timer0 + max volume to L and R

			REG_DMA1SAD	  = (unsigned long)&Soundbuffer;
			REG_DMA1DAD=0x040000a0; 					//write to FIFO A address
			REG_DMA1CNT_H=0xf600;						//dma control: DMA enabled+ start on FIFO+32bit+repeat
		}
		iSamples = 0;
	}
}

//	--------------------------------------------------------------------------------------------
//	callback-function - timer1-interrupt	(called 50 times/s)
//	used to control playback-speed of song!
//	--------------------------------------------------------------------------------------------
void	C64Play() {
		// 0		= value for akku, xreg and yreg (used to initialize .psid-song)
		// play 	= address in c64-memory where to jump
		C64Init(0,play);
}

void (*IntrTable[])()={
  VBlanc,     //v-blank 	(for background movement)
  0,          //h-blank
  0,          //vcount
  0,          //timer0
  C64Play,	  //timer1 		(50hz)
  0,          //timer2
  0,          //timer3
  0,          //serial
  0,          //dma0
  DMA1IRQ,    //dma1		(for soundbuffer-swapping)
  0,          //dma2
  0,          //dma3
  0,          //key
  0,           //cart
};

void AgbMain(void){
  unsigned counter,x,y;
  unsigned tileCounter;


  REG_DISPCNT=0x0080;

  //First init palette
  for(counter=0;counter<256;counter++){
    BG_PALETTE[counter]=master_Palette[counter];
    OBJ_PALETTE[counter]=master_Palette[counter];
  }

  //  clear OAM
  for(counter=0;counter<128;counter++){
  	OAM[counter].Attrib0 	= 160;
	OAM[counter].Attrib1 	= 0;
	OAM[counter].Attrib2 	= 0;
	OAM[counter].RotateScale= 0;
  }

  for(tileCounter=0;tileCounter<32768;tileCounter++){
	  SCREEN_16K[tileCounter] = denosis_Tiles[tileCounter];
  }


  //  fill screen 3 (mauer = 64*64)
  for(x=0;x<32;x++)
    for(y=0;y<32;y++) {
	  // tile mauer
      MAP_3[x+(y*32)] =     (x & 7)+ ((y & 7)*32)	;
	  // tile blauesding
      MAP_2[x+(y*32)] = 8 + (x & 7)+ ((y & 7)*32)	;
	  // tile rotesding
      MAP_1[x+(y*32)] = 16+ (x & 7)+ ((y & 7)*32)	;
      // denys roxx-map
      MAP_0[x+(y*32)] = 16*32;
      if ( (y<8) && (x>3) )
      MAP_0[x+(y*32)] = (8*32) + x + ((y & 7)*32)	-3;

  }

  vxDisplayText();

  //Enable and set up the bg
  // 084 = 010000100;
  REG_BG0CNT=0x0084;
  REG_BG0VOFS=0;
  REG_BG0HOFS=0;

  // 184 = 110000100
  REG_BG1CNT=0x0184;
  REG_BG1VOFS=0;
  REG_BG1HOFS=0;

  REG_BG2CNT=0x0284;
  REG_BG2VOFS=0;
  REG_BG2HOFS=0;

  REG_BG3CNT=0x0384;
  REG_BG3VOFS=0;
  REG_BG3HOFS=0;

  // fade in background-grafics (from white to normal color)
  REG_BLDMOD=0x008f;
  REG_COLEY2=0;
  REG_DISPCNT=	H_BLANK_OAM|OBJ_ENABLE|BG3_ENABLE|BG2_ENABLE|BG1_ENABLE|BG0_ENABLE|OBJ_MAP_1D;
  for(counter=0;counter<16;counter++){
    REG_COLEY2=16-counter;
    // wait 3 vblancs:
   	while(REG_VCOUNT!=160){}
   	while(REG_VCOUNT!=161){}
   	while(REG_VCOUNT!=160){}
   	while(REG_VCOUNT!=161){}
   	while(REG_VCOUNT!=160){}
   	while(REG_VCOUNT!=161){}
  }

  // copy .psid file in c64-memory and start playback
  CopySID();
  // now init song:
  C64Init(start-1,init);
  //  start DMA
  DMAInit();

  // enable vblanc-irq
  REG_IME= 0x00;
  REG_DISPSTAT = REG_DISPSTAT|0x08;
  REG_IE = REG_IE|0x01;				// activate vblanc-irq for background-movement
  REG_IME= 0x01;
  while(1){

	  // soundbuffer-fillroutine
	  if(iFillBuffer!=-1)	{
		  Fill_Sound_Buffer(iFillBuffer<<8);
		  iFillBuffer=-1;
	  }
	  if (iAktiveBuffer==150)
	  	break;
  }

}
