        .ALIGN
        .ARM
        .SECTION 	.ewram,"ax",%progbits
	.GLOBAL		InitDMAPlay
	.GLOBAL		Soundbuffer
	.GLOBAL		Fill_Sound_Buffer
	.GLOBAL		Memory
@----------------------------------------------------------------------------
@		CBM64_SID6581_Player
@		by Denys Bogatz, based on my !MOS6581 for RiscOS
@----------------------------------------------------------------------------

.ascii	"MOS6581/MOS6502-Emulator"
.ALIGN

@----------------------------------------------------------------------------
.equ	REG_TM0CNT	,   	0x0000	@ 32 Timer 0
.equ	REG_TM0CNT_L	,	0x0000	@ 16 Timer 0 count value
.equ	REG_TM0CNT_H	,   	0x0002	@ 16 Timer 0 Control
.equ	REG_TM1CNT_L	,   	0x0004	@ 16 Timer 2 count value
.equ	REG_TM1CNT_H	,   	0x0006	@ 16 Timer 2 control

.equ	REG_SOUNDCNT	,   	0x0000	@ 32
.equ	REG_SOUNDCNT_L	,	0x0000	@ 16 DMG sound control
.equ	REG_SOUNDCNT_H	,	0x0002	@ 16 Direct sound control
.equ	REG_SOUNDCNT_X	,	0x0004	@ 16 Extended sound control

.equ	REG_DMA1SAD	,	0x00BC	@ 32 DMA1 Source Address
.equ	REG_DMA1DAD	,	0x00C0	@ 32 DMA1 Desination Address
.equ	REG_DMA1CNT	,	0x00C4	@ 32 DMA1 Control (Amount)
.equ	REG_DMA1CNT_L	,	0x00C4	@ 16 DMA1 Control Low Value
.equ	REG_DMA1CNT_H	,	0x00C6	@ 16 DMA1 Control High Value
.equ	REG_IE		,	0x0000	@ 16 Interrupt Enable
.equ	REG_IME		,	0x0008	@ 16 Interrupt Master Enable


InitDMAPlay:
	STMFD	R13!,{R0-R12,R14}	@ preserve registers
	LDMFD	R13!,{R0-R12,r14}
	bx	lr			@ uses "bx lr" rather than "movs pc,lr" ('coz thumb/arm-mixin')
@----------------------------------------------------------------------------

@----------------------------------------------------------------------------
Maske:		.word	0x0003FFFF
ADSR_64:	.byte	0,0,0,0		@ reflects state of ADSR-Register
VOLUMES:	.word	0,0,0,0		@ volumes for voices (0-0x10000)
@----------------------------------------------------------------------------

SoundbufferPosition:
	.word	0x0,0
Screenbuffer:	
	.word 	0x06000000
	
Fill_Sound_Buffer:
	@ r0 = offset into soundbuffer 
	STMFD	R13!,{R0-R12,R14}	@ preserve registers
	adrl	r1,Soundbuffer
	add	r0,r0,r1
	str	r0,SoundbufferPosition

	@ copy sid for debugging
	adrl	r0,Sidcopy
	adrl	r1,Memory+0xd400
	ldmia	r1,{r2,r3,r4,r5,r6,r7,r8,r9}	
	stmia	r0,{r2,r3,r4,r5,r6,r7,r8,r9}	
	
	@ calc voice 1
	mov	r7,#0
	adrl	r12,Workingbuffer
	add	r10,r12,#0x0100
	bl	CalcVoice
	@ calc voice 2
	mov	r7,#1
	adrl	r12,Workingbuffer+0x100
	add	r10,r12,#0x0100
	bl	CalcVoice
	@ calc voice 3
	mov	r7,#2
	adrl	r12,Workingbuffer+0x200
	add	r10,r12,#0x0100
	bl	CalcVoice

	@ filter to be done!
@	bl	CALCFILTER
	



	@ -------------------------------------------------------
	@ mix voices together:
	@ -------------------------------------------------------
	ldr	r0,SoundbufferPosition
	adrl	r1,Workingbuffer
	add	r2,r1,#0x100
	
	adrl	r6,Memory+0xd400	@ base of sid
mixin:
	add	r4,r1,#0x100
	add	r5,r1,#0x200
	ldrsb	r3,[r1],#1		@ voice1
	ldrsb	r4,[r4]			@ voice2
	ldrsb	r5,[r5]			@ voice3

	@	add the voices 
	add	r3,r3,r4
	add	r3,r3,r5
	mov	r3,r3,asr#2

	@	write to playbackbuffer
	strb	r3,[r0],#1
	cmp	r1,r2
	blt 	mixin
	@ -------------------------------------------------------

	LDMFD	R13!,{R0-R12,r14}
	bx	lr			@ uses "bx lr" rather than "movs pc,lr" ('coz thumb/arm-mixin')
	@ -------------------------------------------------------
	
@----------------------------------------------------------------------------
@ *-----------------------------------------------------*
@ *             SOUND BUFFER FILL ROUTINES              *
@ *-----------------------------------------------------*
@----------------------------------------------------------------------------
CalcVoice:
	@ -------------------------------------------------------
	@ on input:
	@ r7  = voice (0-2)
	@ r12 = pointer to working-buffer
	@ r10 = pointer to END of working buffer
	@ -------------------------------------------------------

	STMFD	R13!,{R14}
	
	ADRL	R4,Memory		@ base of C64-Memory
	ADD	R4,R4,#0xD400		@ base of C64-SID
	ADD	R4,R4,R7,LSL#3		
	SUB	R4,R4,R7		@ which voice [$d400, $d407, $d40e]
	
	LDRB	R0,[R4,#4]		@ Soundreg
	LDR	R8,Maske
	ADR	R5,ADSR_64		@ Pointer to table with status of sid-ADSR

	@ ---- 	when bit in reg.4 not set we MUST go in release-state
	TST	R0,#1
	BEQ	Release

	@ ----	modify volume by ADSR
	@	use our ADSR-Table which reflects the state of adsr
	LDRB	R1,[R5,R7]

	CMP	R1,#0
	BEQ	Attack
	
	CMP	R1,#1
	BEQ	Decay

	@ ----	   #2 
	@	is sustain - just hold volume at current level and do nothing!
	
	CMP	R1,#3
	BGE	Attack

ChkLoop:

	@ ----	read volume for this voice:	(remx: overall volume not added yet)
	ADRl	R5,VOLUMES
	ldr	r9,[r5,r7,lsl#2]
	

	@ ----	test here for waveforms:
	AND	R11,R0,#0xF6		@ upper nibble is waveform; lower is for ring/sync

	LDRB	R0,[R4,#0]
	LDRB	R1,[R4,#1]
	ADD	R0,R0,R1,LSL#8		@ R0 = Frequency of Voice!
	
	ADRL	R2,FrqBuff		
	LDR	R3,[R2,R7,LSL#2]	@ R3 = last freqency-value

	@ ----	some stuff for synchronisation 
	@	(we have reset our oscilator when the other voice overflows)
	adrl	r2,Memory
	adrl	r8,SyncHlp
	ldr	r8,[r8,r7,lsl#2]
	add	r2,r2,r8
	ldrb	r1,[r2,#1]	@ low  frequency
	ldrb	r2,[r2,#1]	@ high frequency
	add	r1,r1,r2,lsl#8	@ frequency for modulation	-> R1
	@ ----	get last frequency value
	adrl	r8,SyncVoc
	ldr	r8,[r8,r7,lsl#2]
	ADRL	R2,FrqBuff		
	LDR	R2,[R2,R8,LSL#2]@ R2 = last freqency-value for the other voice

	LDR	R8,Maske

	@ ----	checkin' waveform
	tst	R11,#0x10
	tstne	R11,#0x04		@ Ringmodulation? (always played as triangle coz i woz lazy)
	bne	TriangleModulated	@ modulated triangle placed here (prevends playing "normal" triangle)

	@ ----	because of some "unsupported" waveform (like "trisaw") 
	@	I use "tst" instead of "cmp"... so "trisaw" (which is 0x30) get played as "normal triangle"...
	and	r5,r11,#0xf0
	
	tst	R5,#0x10	
	bne	Triangle
	
	tst	R5,#0x20
	bne	Sawtooth
	
	tst	R5,#0x40
	bne	Square
	
	tst	R5,#0x50		@ trisquare
	bne	TriSquare		@ remx: not sure about the correct calculation (may be better having those waveforms smpled...)

	tst	R5,#0x80
	bne	Noise			@ humm... I think the frequency for the noise-table does not fit...
	
	@ ----	unknown/unsupported waveform or 0x00 (mute):
	B	ClrFill

@----------------------------------------------------------------------------

Attack:
	STMFD	R13!,{R0-R12,R14}	@ preserve registers

	ADRL	r3,ADSR_64
	ADRL	r4,Memory
	add	r4,r4,#0xd400		@ SID-basepointer
	ADRL	r5,VOLUMES		@ pointer to volume-table
	ADRL	r6,AttackRate		@ table with "speed"-values for attack
	
	add	r4,r4,r7,lsl#3
	sub	r4,r4,r7		@ add voice*7 to SID-pointer
	
	ldr	r1,[r5,r7,lsl#2]	@ get current volume in r1
	ldrb	r2,[r4,#5]		@ read value from AD-Register
	mov	r2,r2,lsr#4		@ r2 now attack-value
	ldr	r2,[r6,r2,lsl#2]	@ r2 now "speed" for increasing volume
	
	add	r1,r1,r2		@ increase volume by speed
	
	cmp	r1,#0x10000		@ max volume reached?
	movge	r1,#0x10000		@ keep max. value!
	movge	r2,#1			@ initiate decay
	strgeb	r2,[r3,r7]		@ must become decay-phase next time!
	
	str	r1,[r5,r7,lsl#2]	@ write back new volume
	
	LDMFD	R13!,{R0-R12,r14}
	b ChkLoop
@	bx	lr			@ uses "bx lr" rather than "movs pc,lr" ('coz thumb/arm-mixin')

Decay:
	STMFD	R13!,{R0-R12,R14}	@ preserve registers

	ADRL	r3,ADSR_64
	ADRL	r4,Memory
	add	r4,r4,#0xd400		@ SID-basepointer
	ADRL	r5,VOLUMES		@ pointer to volume-table
	ADRL	r6,DecayRate		@ table with "speed"-values for decay
	
	add	r4,r4,r7,lsl#3
	sub	r4,r4,r7		@ add voice*7 to SID-pointer
	
	ldr	r1,[r5,r7,lsl#2]	@ get current volume in r1
	
	ldrb	r2,[r4,#5]		@ read value from AD-Register
	and	r2,r2,#0x0f		@ r2 now decay-value
	ldr	r2,[r6,r2,lsl#2]	@ r2 now "speed" for decreasing volume
	
	ldrb	r0,[r4,#6]		@ read value from SR-Register
	mov	r0,r0,lsr#4		@ r0 is volume we've to reach
	orr	r0,r0,r0,lsl#4
	orr	r0,r0,r0,lsl#8		@ range now 0x0000 - 0xffff (this volume must be reached)
	
	
	sub	r1,r1,r2		@ decrease volume by speed

	cmp	r1,r0			@ sustain-level reached?
	movle	r1,r0			@ keep sustain-level
	movle	r2,#2			@ initiate sustain
	strleb	r2,[r3,r7]		
	
	str	r1,[r5,r7,lsl#2]	@ write back new volume
	
	LDMFD	R13!,{R0-R12,r14}
	b ChkLoop
@	bx	lr			@ uses "bx lr" rather than "movs pc,lr" ('coz thumb/arm-mixin')

Release:
	STMFD	R13!,{R0-R12,R14}	@ preserve registers

	ADRL	r3,ADSR_64
	ADRL	r4,Memory
	add	r4,r4,#0xd400		@ SID-basepointer
	ADRL	r5,VOLUMES		@ pointer to volume-table
	ADRL	r6,ReleaseRate		@ table with "speed"-values for release
	
	add	r4,r4,r7,lsl#3
	sub	r4,r4,r7		@ add voice*7 to SID-pointer
	
	ldr	r1,[r5,r7,lsl#2]	@ get current volume in r1
	
	ldrb	r2,[r4,#6]		@ read value from SR-Register
	and	r2,r2,#0x0f		@ r2 now decay-value
	ldr	r2,[r6,r2,lsl#2]	@ r2 now "speed" for decreasing volume (release)
	
	sub	r1,r1,r2		@ decrease volume by speed

	cmp	r1,#0			@ min. volume reached?
	movle	r1,#0			@ keep it 0!
	movle	r2,#3			@ keep it in release-phase!
	strleb	r2,[r3,r7]		
	
	str	r1,[r5,r7,lsl#2]	@ write back new volume
	
	LDMFD	R13!,{R0-R12,r14}
	b ChkLoop
@	bx	lr			@ uses "bx lr" rather than "movs pc,lr" ('coz thumb/arm-mixin')


@----------------------------------------------------------------------------

Triangle:
	@ R0 frequency  of aktive voice
	@ R1 frequency for modulation
	@ R2 last value of frequency for modulation
	@ R3 last value of aktive frequency
	@ R8 = 0x0003FFFF
	@ R11= waveform register (need to check for synchronisation)
	
	@ ----	sync-check:
	add	r2,r2,r1
	tst	r11,#0x02
	tstne	r2,#0x40000	@ syncvoice overflows!
	movne	r3,#0		@ must then reset this voice
	movne	r2,#0		@ ...and startup counting
	
	ADD	R3,R3,R0
	TST	R3,#0x00020000
	MOVEQ	R5,R3,LSL#1
	MVNNE	R5,R3,LSL#1
	
	AND	R5,R5,R8

	mvn	R5,R5,LSR#10	@ 0x00-0xff
	
	and	r5,r5,#0xff
	sub	r5,r5,#0x80
	
	mul	r5,r9,r5	@ patch volume 
	mov	r5,r5,asr#16

	
	STRB	R5,[R12],#1
	
	CMP	R12,R10
	BLT	Triangle
	B	EndFill

@----------------------------------------------------------------------------

TriangleModulated:
	@ R0 frequency  of voice
	@ R3 last value of frequency
	@ R7 = voice 0-2
	@ R8 = 0x0003FFFF
	
	@ ----	get frequency for the voice we got to be modulated
	adrl	r4,Memory
	adrl	r5,SyncHlp
	ldr	r5,[r5,r7,lsl#2]
	add	r4,r4,r5
	ldrb	r1,[r4,#1]	@ low  frequency
	ldrb	r2,[r4,#1]	@ high frequency
	add	r1,r1,r2,lsl#8	@ frequency for modulation	-> R1
	@ ----	get last frequency value
	adrl	r5,SyncVoc
	ldr	r5,[r5,r7,lsl#2]
	ADRL	R2,FrqBuff		
	LDR	R2,[R2,R5,LSL#2]@ R2 = last freqency-value for the other voice

RingModulation:	
	@ ----	"normal" triangle processing:
	ADD	R3,R3,R0
	TST	R3,#0x00020000
	MOVEQ	R5,R3,LSL#1
	MVNNE	R5,R3,LSL#1
	
	AND	R5,R5,R8
	mvn	R5,R5,LSR#10	@ 0x00-0xff
	and	r5,r5,#0xff

	@ ----	check MSB - then "reverse" triangle if needed:
	@ ----	(MSB of the "other" voice affects this waveform			)
	@ ----	(triangle will be f(x) = x on MSB = 0; f(x) = 255-x on MSB = 1	)
	ADD	R2,R2,R1
	TST	R2,#0x00020000
	RSBNE	r5,r5,#0xff
	
	sub	r5,r5,#0x80	@ make waveform signed
	
	mul	r5,r9,r5	@ patch volume 
	mov	r5,r5,asr#16

	
	STRB	R5,[R12],#1
	
	CMP	R12,R10
	BLT	RingModulation
	B	EndFill

@----------------------------------------------------------------------------

Sawtooth:
	@ R0 frequency  of aktive voice
	@ R1 frequency for modulation
	@ R2 last value of frequency for modulation
	@ R3 last value of aktive frequency
	@ R8 = 0x0003FFFF
	@ R11= waveform register (need to check for synchronisation)
	
	@ ----	sync-check:
	add	r2,r2,r1
	tst	r11,#0x02
	tstne	r2,#0x40000	@ syncvoice overflows!
	movne	r3,#0		@ must then reset this voice
	movne	r2,#0		@ ...and startup counting


	ADD	R3,R3,R0
	AND	R3,R3,R8
	
	MVN	R5,R3,LSR#10
	and	r5,r5,#0xff
	sub	r5,r5,#0x80

	mul	r5,r9,r5	@ patch volume
	mov	r5,r5,asr#16

	STRB	R5,[R12],#1
	CMP	R12,R10
	BLT	Sawtooth
	B	EndFill

@----------------------------------------------------------------------------

Square:
	@ ---- 	get pulsewidth
	LDRB	R5,[R4,#2]
	LDRB	R6,[R4,#3]
	AND	r6,r6,#0x0f
	ADD	R6,R5,R6,LSL#8	@ 12bits for pulsewidth!
	@ ----	remx: having a pulse-width of 2048 (that means 50/50) and 2048<<6 gives 0x20000 (~half of 0x3ffff)
@	mov 	r6,r6,lsl#6	@ now at 18 bits as our max. frequence is... (max. 0x3ffff ~ 18bits)

FillSquare:
	@ R0 frequency  of aktive voice
	@ R1 frequency for modulation
	@ R2 last value of frequency for modulation
	@ R3 last value of aktive frequency
	@ R6 pulse-width	--------------------- <<< extra
	@ R8 = 0x0003FFFF
	@ R11= waveform register (need to check for synchronisation)
	
	@ ----	sync-check:
	add	r2,r2,r1
	tst	r11,#0x02
	tstne	r2,#0x40000	@ syncvoice overflows!
	movne	r3,#0		@ must then reset this voice
	movne	r2,#0		@ ...and startup counting


	ADD	R3,R3,R0
	AND	R3,R3,R8	@ and 0x3ffff
	
@	CMP	R3,R6		@ upper or lower pulse?
cmp r6,r3,lsr#6	
	MOVGE	R5,#0xff
	MOVLT	R5,#0x00
	
	sub	r5,r5,#0x80
	
	mul	r5,r9,r5	@ patch volume
	mov	r5,r5,asr#16

	STRB	R5,[R12],#1
	CMP	R12,R10
	BLT	FillSquare

	B	EndFill

@----------------------------------------------------------------------------

	@ ----	must be a mix of triangle and square!?
TriSquare:
	@ ---- 	get pulsewidth
	LDRB	R5,[R4,#2]
	LDRB	R6,[R4,#3]
	and	r6,r6,#0x0f
	ADD	R6,R5,R6,LSL#8	@ 12bits for pulsewidth!
	@ ----	remx: having a pulse-width of 2048 (that means 50/50) and 2048<<6 gives 0x20000 (~half of 0x3ffff)
	mov 	r6,r6,lsl#6	@ now at 18 bits as our max. frequence is... (max. 0x3ffff ~ 18bits)
FillTriSquare:
	@ R0 frequency  of aktive voice
	@ R1 frequency for modulation
	@ R2 last value of frequency for modulation
	@ R3 last value of aktive frequency
	@ R6 pulse width	---------------------------- <<<
	@ R8 = 0x0003FFFF
	@ R11= waveform register (need to check for synchronisation)
	
	@ ----	sync-check:
	add	r2,r2,r1
	tst	r11,#0x02
	tstne	r2,#0x40000	@ syncvoice overflows!
	movne	r3,#0		@ must then reset this voice
	movne	r2,#0		@ ...and startup counting


	ADD	R3,R3,R0
	TST	R3,#0x00020000
	MOVEQ	R5,R3,LSL#1
	MVNNE	R5,R3,LSL#1
	
	AND	R5,R5,R8

	mvn	R5,R5,LSR#10	@ 0x00-0xff
	
	and	r5,r5,#0xff
	sub	r5,r5,#0x80
	
	mul	r5,r9,r5	@ patch volume 
	mov	r5,r5,asr#16

	@ ----	check pulse-width and set voice to mute when MSB set - not sure is this is correct...
	CMP	R3,R6		
	MOVGE	r5,#0
	
	STRB	R5,[R12],#1
	CMP	R12,R10
	BLT	FillTriSquare

	B	EndFill

@----------------------------------------------------------------------------

Noise:	ADRL	R1,NoiseTab
	LDR	R6,NoCount
	ADD	R6,R6,#0x85
	STR	R6,NoCount

FillNoise:
	ADD	R3,R3,R0
	ADD	R5,R6,R3,LSR#7
	AND	R5,R5,R8
	LDRB	R5,[R1,R5,LSR#7]

	mul	r5,r9,r5	@ patch volume
	mov	r5,r5,asr#16

	STRB	R5,[R12],#1
	CMP	R12,R10
	BLT	FillNoise

	B	EndFill

NoCount:.word	0x00000000

@----------------------------------------------------------------------------
ClrFill:MOV	R5,#0x00

ClrLoop:STRB	R5,[R12],#1
	CMP	R12,R10
	BLT	ClrLoop

	B	ClrEnd

@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
@----------------------------------------------------------------------------

EndFill:
	@ ----	NOTE: get sure you've the last frequency-value in R3 when you go here...
	ADR	R2,FrqBuff	
	STR	R3,[R2,R7,LSL#2]	@ remember last frequency-value!

ClrEnd:	LDMFD	R13!,{PC}

@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
FrqBuff:.word	0,0,0
SyncFrq:.word	0
SyncAdd:.word	0
SyncHlp:.word 	0x0000D40E, 0x0000D400, 0x0000D407	@ also used for ringmodulation!
SyncVoc:.word	2*4, 0*4, 1*4
@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
.ascii	"copy of sid:"
.ALIGN
Sidcopy:
.word	0,0,0,0,0,0,0,0	@ 32 bytes

@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
@----------------------------------------------------------------------------
.ascii	"SOUNDBUFFER:"
.ALIGN
Soundbuffer:		@ 16 buffers (reduces "clicks" for dma-resets!)
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080

.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080
.hword	0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080,0x8080


.ascii	"WORKINGBUFFER"
.ALIGN
Workingbuffer:		@ buffer for 3 voices (each 0x0100 bytes)
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00

.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00

.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
.word	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00

.ascii	"AttTab"
.ALIGN
@
@	timing for ADSR
@	Attack(15) will take 8 seconds from volume 0..15
@	I'm not sure if attack(15) also takes 8s from e.g. 10..15
@	(so I'll take just one "speedtab" instead of 16 - considering "attack"-speed will always the same...)
@
@  +-----+------------------------+--------------------------------+
@  |VALUE|ATTACK RATE (TIME/CYCLE)| DECAY/RELEASE RATE (TIME/CYCLE)|
@  +-----+------------------------+--------------------------------+
@  |  0  |           2 ms         |               6 ms             |
@  |  1  |           8 ms         |              24 ms             |
@  |  2  |          16 ms         |              48 ms             |
@  |  3  |          24 ms         |              72 ms             |
@  |  4  |          38 ms         |             114 ms             |
@  |  5  |          56 ms         |             168 ms             |
@  |  6  |          68 ms         |             204 ms             |
@  |  7  |          80 ms         |             240 ms             |
@  |  8  |         100 ms         |             300 ms             |
@  |  9  |         250 ms         |             750 ms             |
@  | 10  |         500 ms         |             1.5 s              |
@  | 11  |         800 ms         |             2.4 s              |
@  | 12  |           1 s          |               3 s              |
@  | 13  |           3 s          |               9 s              |
@  | 14  |           5 s          |              15 s              |
@  | 15  |           8 s          |              24 s              |
@  +-----+------------------------+--------------------------------+

@	due to our soundbuffer which holds 256 bytes we've to increase/decrease our volume
@	by the following values:
@	(16khz playback = 16384bytes/s = 64 buffers/s played)
@	value of 0x10000 means full volume -> (waveform*thisvolume)>>16
AttackRate:
.word	512000	@(0x10000 / 64)	/0.002	@ full volume next quarter of buffer		(nyi)
.word	128000	@(0x10000 / 64)	/0.008	@ full volume within next half buffer		(nyi)
.word	 64000	@(0x10000 / 64)	/0.016	@ approx. full volume within next buffer
.word	 42666	@(0x10000 / 64)	/0.024
.word	 26947	@(0x10000 / 64)	/0.038
.word	 18285	@(0x10000 / 64)	/0.056
.word	 15058	@(0x10000 / 64)	/0.068
.word	 12800	@(0x10000 / 64)	/0.08
.word	 10240	@(0x10000 / 64)	/0.1
.word	  4096	@(0x10000 / 64)	/0.25
.word	  2048	@(0x10000 / 64)	/0.5
.word	  1280	@(0x10000 / 64)	/0.8
.word	  1024	@(0x10000 / 64)	/1
.word	   341	@(0x10000 / 64)	/3
.word	   205	@(0x10000 / 64)	/5
.word	   128	@(0x10000 / 64)	/8

DecayRate:
ReleaseRate:
.word	170666	@(0x10000 / 64)	/0.006
.word	 42666	@(0x10000 / 64)	/0.024
.word	 41333	@(0x10000 / 64)	/0.048
.word	 14222	@(0x10000 / 64)	/0.072
.word	  8982	@(0x10000 / 64)	/0.114
.word	  6095	@(0x10000 / 64)	/0.168
.word	  5019	@(0x10000 / 64)	/0.204
.word	  4266	@(0x10000 / 64)	/0.240
.word	  3413	@(0x10000 / 64)	/0.3
.word	  1365	@(0x10000 / 64)	/0.75
.word	   682	@(0x10000 / 64)	/1.5
.word	   426	@(0x10000 / 64)	/2.4
.word	   341	@(0x10000 / 64)	/3
.word	   114	@(0x10000 / 64)	/9
.word	    68	@(0x10000 / 64)	/15
.word	    43	@(0x10000 / 64)	/24

.ascii	"NOIS"
NoiseTab:
.ALIGN
.byte 0x7A,0xED,0x3A,0xC1,0xEB,0x45,0x9E,0x75,0x2D,0xE2
.byte 0x3F,0x42,0x56,0x9E,0x6B,0x99,0xB8,0xD7,0x9E,0x67
.byte 0xB8,0x70,0xFF,0x9E,0xAE,0xD9,0xAA,0x19,0xCF,0xC8
.byte 0xC1,0xE4,0x02,0x3F,0xDF,0x75,0x86,0x48,0xAA,0xB1
.byte 0x05,0x9E,0x75,0xB1,0x48,0x62,0x6B,0x06,0x10,0xAA
.byte 0x18,0xDE,0x3D,0x48,0xF9,0xBE,0x75,0x9E,0x7A,0x56
.byte 0x2B,0x7F,0x42,0x03,0xED,0xC1,0x31,0xBB,0x5A,0xE4
.byte 0xE7,0xAA,0xF2,0x7A,0x90,0x52,0x45,0xD7,0x67,0x7F
.byte 0xF7,0x99,0xF9,0xE3,0x7F,0xA6,0x09,0x8B,0xC3,0x35
.byte 0x70,0x9E,0x56,0xCF,0xD3,0x42,0x75,0xDC,0xDC,0xA2
.byte 0x3F,0x22,0xE9,0x48,0x2D,0xCF,0x01,0x70,0x8B,0xAE
.byte 0x81,0x35,0xA2,0xC1,0x13,0x75,0x6B,0x7A,0x13,0x5A
.byte 0x45,0x7F,0x3F,0x09,0xAE,0xE7,0x95,0xA6,0x4F,0x8B
.byte 0x95,0x29,0x35,0xBB,0x48,0x67,0xA6,0x10,0xD9,0xFD
.byte 0xE3,0x26,0x5A,0xA2,0x3F,0x45,0xDF,0x52,0x5A,0x27
.byte 0x48,0x38,0xF5,0xBE,0x35,0xA6,0xB1,0x95,0xDE,0x67
.byte 0x06,0xC6,0x7A,0xCB,0x5A,0x4C,0x35,0xA6,0x7A,0x5E
.byte 0xF1,0xE4,0x04,0x6B,0x0D,0xB8,0x10,0x3D,0x52,0x45
.byte 0x95,0x86,0xAE,0xFA,0x81,0x5E,0xAA,0xFD,0x48,0x4F
.byte 0x45,0xDC,0x90,0xC3,0xF4,0xAA,0x4F,0xC3,0x81,0x95
.byte 0xA6,0xCD,0xB1,0x5A,0xFD,0x81,0xF1,0x1C,0x48,0xED
.byte 0xEF,0xD3,0x2F,0x05,0x86,0x95,0x7F,0x09,0x95,0xEE
.byte 0x75,0x1D,0x48,0xBB,0x19,0x1C,0x9E,0x1D,0xAE,0x8B
.byte 0x11,0x15,0x45,0x31,0xF4,0x38,0x7F,0x81,0xDA,0x4C
.byte 0x3F,0xFC,0x95,0xE9,0xAE,0xCD,0x52,0x70,0x95,0x29
.byte 0x62,0x4F,0x42,0x86,0x1A,0x4C,0x6B,0xA6,0xE3,0xF4
.byte 0x62,0x95,0xEE,0xEE,0xFC,0x2B,0x56,0x33,0x10,0x19
.byte 0xE3,0xED,0x4C,0xB1,0x1D,0xB1,0xF7,0x0E,0x86,0x62
.byte 0x52,0x52,0xD3,0x0C,0xDF,0x24,0x75,0xEF,0x90,0x06
.byte 0xE2,0x62,0xEE,0x52,0x7A,0xC3,0x99,0xD3,0xCD,0xB1
.byte 0xC8,0x24,0x0A,0x81,0x7A,0x5A,0xCF,0xE0,0x45,0x8B
.byte 0x7F,0x4F,0x99,0x62,0xCF,0xA6,0x48,0xE6,0x75,0x7F
.byte 0x29,0x9E,0xB4,0x21,0x6B,0xA6,0xB4,0x13,0x7F,0x99
.byte 0x27,0x06,0x90,0xB4,0x48,0x86,0x6B,0x7A,0xBB,0xB1
.byte 0x86,0x9E,0x33,0x62,0x45,0xE7,0xA2,0x70,0xFB,0x14
.byte 0x24,0xED,0x7A,0xB8,0xDA,0x81,0x99,0x81,0x14,0x95
.byte 0x5A,0x14,0xE6,0xEF,0xB8,0x56,0x95,0xD7,0xEF,0x2D
.byte 0x86,0x1E,0x62,0x90,0x67,0x8B,0xC3,0xD1,0x13,0x6B
.byte 0x90,0xC8,0x9E,0xA6,0x7A,0x62,0xF8,0xD9,0x45,0x5A
.byte 0x4C,0xF2,0x3A,0x16,0xAE,0xE7,0x4F,0x56,0x7F,0xD7
.byte 0x1D,0xEF,0xEA,0xAA,0xAA,0xE7,0xD5,0xEC,0x6B,0xB4
.byte 0x6B,0x81,0xA6,0x14,0xAE,0x8B,0x67,0x48,0x27,0x3D
.byte 0x22,0xA6,0x17,0x6B,0x8B,0x70,0x45,0x5E,0x48,0xF0
.byte 0x8B,0xD1,0x3F,0x9E,0xDA,0xFB,0x5A,0x81,0xDA,0x7F
.byte 0x45,0x90,0xCD,0xAE,0x75,0x81,0xE7,0xEF,0x6B,0xFC
.byte 0xAA,0x3A,0x2B,0xC8,0xCF,0x35,0xF8,0xCD,0xD1,0x17
.byte 0xBB,0xB4,0xDE,0x45,0x90,0x6B,0xE7,0x0D,0xED,0x13
.byte 0x10,0xE4,0xF0,0xD1,0xCF,0x62,0xED,0x99,0x09,0xAE
.byte 0x62,0x04,0x05,0x7A,0x62,0x81,0x7F,0x4C,0xF7,0x70
.byte 0x81,0xB1,0x3A,0x95,0x4F,0xC1,0x90,0x81,0x95,0x81
.byte 0xF7,0x4C,0xAA,0x1D,0xC1,0x14,0x1C,0x0E,0xC8,0x0A
.byte 0xB4,0x4C,0x22,0x5A,0x11,0x90,0x04,0xE0,0x7A,0x1A
.byte 0x5A,0xEE,0x4C,0xF6,0xCD,0x90,0xA6,0x86,0xAA,0x33
.byte 0xBE,0xB4,0x13,0x16,0x0C,0xB4,0xBE,0x04,0xD1,0x99
.byte 0xF7,0xD3,0x4C,0x8B,0x15,0xB4,0x2D,0xA2,0x1A,0xEF
.byte 0x7F,0xF2,0xD9,0x38,0x99,0xFE,0x3A,0xD1,0xA6,0x3D
.byte 0xD5,0x29,0x62,0x00,0xD9,0x67,0x95,0xFD,0xB8,0x99
.byte 0x70,0xB1,0x56,0x8B,0xE8,0x04,0xFF,0x81,0xF2,0xD5
.byte 0xAE,0x7A,0x38,0xDF,0x22,0x33,0x52,0x38,0x33,0x4F
.byte 0x24,0xD7,0x4F,0x67,0x3F,0x4F,0xB1,0x95,0xF5,0xC1
.byte 0x95,0x75,0xFA,0xCD,0xAA,0xC8,0x45,0x45,0x7F,0x5E
.byte 0x5A,0x1A,0xFB,0xFF,0x8B,0x99,0x7A,0x95,0x2D,0xC6
.byte 0x29,0xB1,0xCB,0x4C,0x14,0x7A,0x45,0xB4,0xBE,0xA6
.byte 0x22,0xDF,0xBB,0x95,0x12,0x48,0xA6,0x62,0xD5,0x75
.byte 0xEC,0xB1,0x95,0xF2,0x95,0x90,0xEA,0x4C,0x99,0x70
.byte 0xAE,0x81,0x3D,0x5A,0x70,0xD7,0x2B,0xAE,0x95,0xD3
.byte 0x13,0x22,0x75,0xF0,0xB8,0x81,0x33,0xE9,0xE9,0xD3
.byte 0xC1,0x35,0x4F,0x7A,0xA6,0x7F,0xC6,0x02,0x9E,0xEF
.byte 0xE4,0x31,0x81,0x7F,0x13,0x95,0xDA,0xF1,0x3F,0x45
.byte 0xE9,0x29,0x4C,0x6B,0x0B,0xB8,0x29,0xB4,0x42,0x86
.byte 0xE0,0xD5,0xBE,0x8B,0x81,0x5E,0x29,0xE6,0xF4,0x5E
.byte 0x86,0xF3,0x90,0xB8,0xAA,0x2D,0xDC,0x22,0x81,0xC3
.byte 0x45,0x3F,0x7F,0x1E,0x3A,0x48,0x2D,0x7F,0xAA,0x70
.byte 0x90,0x7F,0x8B,0x7F,0x99,0xD7,0xE7,0xF4,0x16,0xAA
.byte 0x56,0xB4,0x2F,0xD3,0xBE,0xF7,0x81,0xCB,0x62,0x5E
.byte 0x7F,0xC8,0xE9,0x27,0x4F,0x13,0x45,0x5E,0x2D,0x5E
.byte 0xB1,0x62,0xDA,0x20,0x86,0x52,0x18,0x7F,0x4C,0x08
.byte 0x45,0x86,0x81,0x27,0x5E,0xB4,0x7A,0xC3,0xC6,0x00
.byte 0x75,0xCF,0x24,0x11,0xF9,0x48,0xC6,0x70,0x81,0xC8
.byte 0x52,0x22,0xED,0xD9,0x1C,0x7A,0xED,0xAA,0x24,0x5E
.byte 0xBB,0x11,0x5E,0x81,0xED,0x03,0xD5,0x6B,0xAE,0xF6
.byte 0x95,0xB8,0x01,0xBE,0xBB,0x7A,0x5E,0x81,0x99,0xFA
.byte 0x95,0x00,0x7F,0xDE,0xFC,0x8B,0xC1,0xCB,0x81,0xE3
.byte 0x62,0x9E,0xB8,0x81,0xC6,0x5E,0xA2,0x62,0x48,0xCD
.byte 0x81,0x5A,0x42,0x08,0xA2,0xAE,0x45,0x5A,0x52,0x27
.byte 0x18,0xEC,0x4F,0xDF,0x4C,0xA2,0xC8,0xE7,0x09,0x9E
.byte 0xC6,0xAA,0xD3,0x9E,0xBE,0xC3,0x7F,0x81,0xC1,0x24
.byte 0x13,0xDE,0xF8,0x2B,0xC3,0xDE,0x9E,0x3A,0x35,0x4C
.byte 0xE8,0xD7,0x38,0x52,0x6B,0x81,0xBB,0x3A,0x20,0x8B
.byte 0xC3,0xCB,0x81,0x2F,0xEC,0xB1,0x7A,0x07,0xC1,0xDC
.byte 0x03,0x75,0x86,0xF2,0xBB,0x06,0x52,0xC3,0xAA,0xDE
.byte 0xA6,0x86,0xEB,0xDE,0x26,0xDA,0xE6,0xFF,0x18,0x52
.byte 0xBB,0x13,0xDF,0xD3,0xD5,0xF4,0x35,0xAA,0x48,0xD3
.byte 0x45,0x62,0xA6,0xA2,0xE6,0x20,0x2D,0x56,0x7F,0x01
.byte 0xD1,0xB8,0x5E,0x45,0x45,0x7F,0x70,0x08,0x4C,0xDA
.byte 0xEF,0xC6,0x81,0x90,0xE6,0xE3,0x16,0x62,0xC1,0x35
.byte 0x1C,0xFC,0x0A,0x81,0xF7,0x5E,0xAA,0xAA,0x4C,0xE0
.byte 0xD7,0x05,0xF6,0xC8,0x35,0xC8,0x75,0xFF,0xFD,0x75
.byte 0x70,0xAE,0xAE,0xFC,0x5E,0x02,0xCD,0xC1,0xDE,0xDC
.byte 0xCD,0xB4,0xEE,0x75,0x4F,0xC8,0xF2,0xC1,0xE8,0x62
.byte 0x90,0x81,0x7F,0x2F,0xED,0xCF,0x13,0x95,0x03,0x09
.byte 0x7F,0x95,0xCD,0x22,0x38,0xD5,0xDC,0x3D,0x18,0xDC
.byte 0x29,0x4F,0x75,0x9E,0x67,0xAE,0xCD,0x8B,0x95,0xBB
.byte 0x18,0xBB,0xE3,0x07,0xD9,0x7F,0xD3,0x56,0xC3,0xD9
.byte 0x70,0x5E,0xF7,0x04,0x70,0x16,0xC6,0xD7,0x45,0xAE
.byte 0x5E,0x9E,0x05,0x5A,0x31,0xEF,0x2F,0x5E,0x20,0xAA
.byte 0x86,0x03,0xC1,0xD3,0x4F,0xEB,0x4F,0xA6,0x29,0xF6
.byte 0x45,0x86,0xB8,0x62,0xBE,0xBE,0x70,0x52,0x0A,0x08
.byte 0x03,0x70,0x48,0x4C,0x95,0x5A,0xFB,0x29,0x9E,0xC8
.byte 0xB4,0x4F,0x56,0x81,0xCD,0x31,0x7A,0x0C,0x5A,0x67
.byte 0x1D,0x3D,0x22,0xBB,0x52,0x2D,0x4C,0x11,0xB4,0x14
.byte 0x7F,0xA6,0x56,0x45,0x04,0xFF,0xE4,0xC3,0x90,0x15
.byte 0xBB,0x27,0x70,0x5A,0x45,0x90,0x5A,0x2D,0xFF,0xB8
.byte 0x2D,0xCB,0x3A,0xE9,0x6B,0x21,0xE3,0xF6,0x62,0x95
.byte 0x4C,0x56,0x19,0xF9,0xB8,0xBE,0x7F,0x3F,0xF3,0x81
.byte 0x0F,0x2D,0x04,0xFC,0xFA,0xF2,0xBB,0x4F,0xA2,0x45
.byte 0xD9,0x21,0x7A,0xAE,0x75,0x29,0x45,0x75,0x09,0x81
.byte 0xF8,0x5A,0x27,0x20,0xEC,0x81,0xB8,0xDC,0x7F,0xDA
.byte 0x1E,0xF9,0x35,0x14,0x1D,0x99,0x62,0x62,0x45,0x75
.byte 0xFF,0xA6,0x42,0x7A,0xF0,0x29,0x7F,0x4C,0x04,0x56
.byte 0x48,0x7F,0xC3,0x00,0xBB,0x3A,0xFB,0x90,0x3A,0x5A
.byte 0xE9,0xD1,0x86,0xBB,0x26,0x15,0x26,0x0C,0x81,0xC6
.byte 0x9E,0xCB,0x3F,0x0F,0xA2,0x10,0x75,0x8B,0x24,0xD7
.byte 0x62,0x56,0x62,0x67,0x5E,0x7F,0xEA,0x86,0xFE,0xD3
.byte 0x19,0x86,0x42,0xEB,0x6B,0x90,0x81,0x08,0xC1,0xFC
.byte 0x62,0xB1,0x7A,0xAA,0x75,0x86,0xBE,0x7F,0xDE,0x38
.byte 0xC6,0x06,0x42,0x16,0x75,0xB1,0xBE,0x9E,0xD1,0xAA
.byte 0x5A,0x5E,0x33,0x99,0x4F,0x67,0x9E,0x48,0x11,0x99
.byte 0xE0,0xD3,0xEB,0x70,0x62,0xCB,0x16,0xC8,0x75,0xD5
.byte 0x38,0x8B,0xCD,0x01,0x2F,0x1C,0xB4,0x1A,0xE2,0x5A
.byte 0x75,0xDE,0x24,0x13,0x5E,0x48,0x14,0xF4,0x67,0x21
.byte 0xA6,0x31,0x5A,0x2D,0x7F,0xC1,0x7A,0xD1,0x75,0xFA
.byte 0xE4,0xD1,0xC6,0xAA,0xED,0x33,0x4F,0xB8,0xFE,0x4C
.byte 0x31,0xAE,0x99,0x01,0xAA,0x9E,0x3F,0x0C,0x20,0x3D
.byte 0x27,0x9E,0xAE,0xC8,0x9E,0x29,0x52,0x31,0x7F,0x3D
.byte 0x4F,0xB8,0xEF,0xED,0x5E,0x52,0xB1,0x38,0x3D,0x7A
.byte 0xCF,0xCF,0x4C,0xB4,0x86,0xF9,0xB1,0xB4,0x99,0xAA
.byte 0xBB,0x45,0x7A,0xCD,0xD9,0x20,0x18,0x2D,0xF5,0x9E
.byte 0x48,0x8B,0x81,0xFB,0x31,0x6B,0x4C,0x95,0x70,0x81
.byte 0xE9,0x81,0x0B,0xA6,0x81,0xFB,0x6B,0x05,0xB8,0xD1
.byte 0xEA,0x2B,0xF3,0x14,0xBE,0xB1,0xFB,0xF3,0xE3,0x06
.byte 0x5E,0x2D,0x2D,0xF0,0xAE,0x75,0xF0,0xCF,0xE2,0xCD
.byte 0xD9,0xF4,0xBB,0xA2,0x1D,0x42,0x2D,0x29,0x7F,0x62
.byte 0x38,0x70,0x31,0xEF,0x29,0x95,0x52,0x86,0xB1,0x2B
.byte 0xF6,0x86,0x21,0xCB,0xFE,0xF7,0x4F,0x05,0x42,0xC6
.byte 0xCD,0x81,0x5E,0x7A,0x1E,0x7A,0xBB,0x62,0x15,0xCD
.byte 0x67,0x1A,0x6B,0xF7,0x81,0x21,0x8B,0xAA,0x04,0x03
.byte 0xF0,0x7F,0x2F,0x4F,0xB1,0x6B,0x13,0x2B,0xEA,0xBB
.byte 0xA6,0xC1,0x4C,0xAE,0xA2,0xAA,0x1C,0xF8,0x70,0xA2
.byte 0x9E,0x5A,0x86,0x99,0xD7,0x5E,0xA6,0xBE,0x52,0xA6
.byte 0x1A,0x24,0x2D,0x2B,0xE4,0x5E,0x70,0x4C,0x95,0x7F
.byte 0xA6,0xD7,0xFA,0x75,0xB1,0x1D,0xF2,0xDA,0x7F,0xED
.byte 0xFC,0x9E,0xE4,0xAA,0x5A,0x17,0x75,0x52,0x42,0xAE
.byte 0x27,0x38,0x81,0x26,0x04,0x33,0x7F,0x0B,0x2F,0xB4
.byte 0xB1,0x07,0x99,0xDC,0x99,0x12,0xDE,0x75,0xBE,0xFE
.byte 0x0A,0x81,0x70,0x8B,0x0A,0x08,0x95,0x45,0xBB,0xF0
.byte 0xDA,0x0E,0x9E,0x7A,0xE8,0xC1,0xC1,0xA2,0x2B,0xD7
.byte 0x8B,0xE3,0xEB,0x90,0xB1,0xAE,0x7A,0xBB,0x90,0xB8
.byte 0x67,0xFF,0x95,0x45,0x5E,0x9E,0xFE,0xE4,0xCF,0x04
.byte 0x35,0xEB,0x0B,0xDC,0x38,0x81,0xFB,0x3A,0x70,0x29
.byte 0x81,0xBB,0xFB,0xAE,0x7F,0xEE,0x42,0xEC,0xA2,0x0C
.byte 0x99,0x21,0x6B,0xBB,0xFC,0x22,0x67,0x81,0xC1,0x75
.byte 0xCB,0x81,0x8B,0x52,0x11,0xA2,0x86,0x56,0x2B,0xB1
.byte 0x5E,0x1A,0x27,0x1D,0xD5,0x5A,0x4F,0xF4,0xB4,0x2F
.byte 0x8B,0x86,0xAA,0x95,0xD3,0x3F,0x95,0x52,0x19,0x62
.byte 0x45,0xE9,0x56,0x7F,0x5A,0x04,0xA2,0x81,0x42,0xEB
.byte 0xDA,0x67,0x70,0x09,0xD7,0x90,0xFB,0x0D,0xD5,0x3F
.byte 0x7F,0x0E,0x18,0x62,0xED,0x5E,0x2D,0xFD,0x9E,0x99
.byte 0xD3,0xEF,0xC6,0x4C,0x7F,0x90,0x9E,0x95,0x99,0x20
.byte 0xF3,0x52,0x52,0xD1,0x5A,0xE3,0x7F,0x95,0xE8,0x3A
.byte 0x2F,0x1A,0xFB,0xE2,0x70,0x75,0xE6,0x81,0x2B,0x75
.byte 0x7F,0xB1,0x86,0x1A,0x8B,0xB4,0x4F,0xAA,0xA2,0x08
.byte 0xB8,0x81,0x3F,0xA6,0xCF,0xE3,0xDA,0x27,0x90,0x26
.byte 0x1C,0xCF,0x7F,0x31,0x3A,0x5A,0x81,0x86,0xFC,0x62
.byte 0x67,0x52,0x3F,0xB4,0xDA,0x19,0xBB,0xCD,0x62,0xB1
.byte 0x86,0x81,0xBB,0x14,0x0D,0x14,0x45,0x62,0xE0,0x3D
.byte 0x8B,0x9E,0x2D,0x1D,0x48,0xC8,0x31,0xD9,0x67,0xC3
.byte 0x10,0xC3,0x8B,0x5E,0x56,0x81,0x75,0x06,0x7F,0x31
.byte 0x81,0x81,0x26,0xDC,0x35,0x99,0xB8,0x9E,0x05,0x13
.byte 0x5A,0x7F,0x48,0x03,0xFB,0x90,0xC6,0x5A,0x62,0xB4
.byte 0x75,0x6B,0x4C,0x0C,0x0D,0x4C,0x24,0x95,0x48,0x95
.byte 0xE6,0x95,0x5A,0x70,0x62,0x35,0x4F,0xDF,0x16,0xD3
.byte 0x07,0xDA,0x70,0x90,0xFE,0xFE,0x75,0xDA,0x48,0xE8
.byte 0x00,0xAE,0x45,0xE9,0xCF,0x90,0xCD,0x08,0x99,0x21
.byte 0x86,0x99,0xBB,0x95,0x0E,0x10,0x67,0x81,0xC8,0x8B
.byte 0xDC,0x70,0x7A,0x29,0x02,0x4F,0xE7,0xE7,0x5E,0x3D
.byte 0x12,0x22,0x1C,0x81,0x13,0x8B,0xB1,0xD3,0x81,0xA6
.byte 0x31,0xCB,0xC6,0xCF,0xE3,0xCF,0x7F,0xB1,0x7A,0x4C
.byte 0xA6,0x86,0x95,0xF8,0x38,0x67,0x86,0xCD,0x52,0x86
.byte 0xD7,0x48,0x00,0x90,0x20,0x81,0x1D,0x6B,0xD5,0x5E
.byte 0x81,0x2B,0x27,0x5A,0x90,0xA2,0xD9,0xAE,0xB1,0xEB
.byte 0xA2,0xC6,0x81,0xDC,0xC1,0xC6,0x1A,0xF9,0xE3,0x48
.byte 0xF8,0xB4,0xE4,0xE8,0x0C,0xEC,0xB4,0x8B,0x70,0xC8
.byte 0x00,0x3D,0xAA,0x1A,0x4C,0xE0,0x8B,0x17,0xD1,0x38
.byte 0x81,0x35,0x0E,0x1A,0x81,0x99,0xB8,0x01,0x81,0x26
.byte 0x9E,0xAA,0x24,0x5E,0x31,0x5E,0x67,0x86,0xF0,0xAA
.byte 0x99,0x86,0x09,0x09,0x8B,0x6B,0x38,0x7F,0x4F,0xFF
.byte 0x3D,0x09,0xFB,0x6B,0xAE,0xD3,0x5A,0x70,0xA6,0xBE
.byte 0x7F,0xCD,0x7F,0x22,0xBB,0x48,0x7F,0x45,0x67,0xD5
.byte 0xDE,0x00,0x2F,0x70,0x81,0x4F,0x7F,0xBE,0x31,0x16
.byte 0x3D,0x2D,0xD1,0xC1,0xCF,0x90,0xB1,0x3F,0x7F,0xE2
.byte 0xAE,0xAE,0x56,0x33,0x15,0xB1,0x7F,0x75,0xAE,0x1A
.byte 0xAA,0x45,0x1A,0xCB,0xB8,0xE8,0x7F,0x4F,0x81,0x20
.byte 0x9E,0x81,0xED,0x5A,0xAE,0x75,0x0C,0xF8,0xD5,0xC6
.byte 0xA6,0xD5,0x81,0x2D,0xEF,0xF3,0xEC,0x2D,0x45,0xED
.byte 0xD7,0x5A,0xE7,0xA6,0x7A,0x09,0xE7,0x8B,0x24,0x70
.byte 0x04,0x7A,0x67,0x52,0x4C,0xCF,0xAE,0xC3,0x6B,0x26
.byte 0xF4,0x99,0x4F,0x99,0x7F,0xD3,0x7F,0xF8,0x95,0xBB
.byte 0x67,0x10,0xAA,0x1E,0x07,0x5E,0xCB,0x0B,0x56,0x38
.byte 0x52,0xCB,0x75,0xCB,0xAE,0x70,0x35,0xF7,0xA2,0xC3
.byte 0x90,0xD3,0x4C,0x4F,0x67,0xDE,0x48,0x9E,0x86,0xD1
.byte 0x2D,0x4C,0xD5,0x52,0xD7,0xFB,0xD7,0x22,0x8B,0x56
.byte 0xAE,0x4C,0x29,0x5E,0x27,0x4F,0xB4,0x6B,0x62,0x3A
.byte 0xBE,0x12,0x3D,0x1C,0x02,0x81,0x27,0x02,0x4C,0xAE
.byte 0xE2,0x17,0xF7,0xCD,0x90,0x7F,0x81,0xE7,0x09,0x6B
.byte 0x75,0x6B,0x17,0x70,0xEA,0x26,0xA2,0x7A,0x81,0xDF
.byte 0x81,0x7A,0x62,0x62,0x48,0xF7,0x2D,0x45,0x04,0xCD
.byte 0x5E,0x33,0x1A,0x3A,0x06,0x52,0x70,0x20,0x3F,0xC1
.byte 0x33,0xDE,0xBB,0x7F,0x86,0x09,0xDC,0xAA,0x62,0xF0
.byte 0x4F,0x90,0xC6,0xE8,0xBB,0x9E,0xED,0x90,0xC8,0xB4
.byte 0xAE,0x95,0x95,0x19,0xC3,0x13,0xB8,0x09,0xBE,0xAE
.byte 0x04,0x45,0x12,0xE9,0x70,0x5A,0xD9,0x29,0x70,0x0C
.byte 0x56,0xFC,0x42,0x56,0xDF,0x90,0x02,0xE2,0x4C,0xB1
.byte 0x0B,0xC8,0x52,0x70,0xF5,0x62,0xE6,0xFF,0x48,0x48
.byte 0x31,0xED,0x7F,0x9E,0x56,0x38,0xA6,0xA6,0xCB,0x7A
.byte 0x3A,0xC6,0x5A,0x81,0x90,0x7F,0x29,0xC1,0xCF,0x1A
.byte 0x90,0x52,0x6B,0x8B,0xF6,0xF5,0xBB,0x52,0xF9,0x29
.byte 0xE0,0x17,0x24,0x5E,0x81,0x2B,0xD9,0x05,0xE0,0xA6
.byte 0x08,0x6B,0xEF,0x75,0xAE,0x29,0xEF,0xCF,0x56,0x7F
.byte 0x62,0xF7,0x3D,0x45,0xC6,0xCF,0xF7,0x56,0x0C,0x2B
.byte 0x38,0x20,0x29,0x7F,0x81,0x70,0x02,0xD9,0x75,0x75
.byte 0xC6,0xD3,0x95,0x09,0x0C,0x4F,0x16,0xA6,0x52,0x52
.byte 0xF4,0x42,0x99,0x2B,0x70,0x99,0x38,0x81,0x7A,0xA2
.byte 0xBB,0xD5,0x86,0xE6,0x4C,0xC8,0x07,0x1C,0xB4,0x81
.byte 0xB4,0x38,0x08,0xD3,0xE2,0x01,0xE4,0x13,0x75,0x4C
.byte 0x4F,0x81,0x31,0xF6,0xBE,0xE3,0x02,0xC1,0x8B,0x81
.byte 0x05,0x62,0xFE,0xCB,0xEC,0xEF,0x01,0x5E,0xE7,0x7A
.byte 0xBE,0x29,0xDC,0xD1,0x18,0xC6,0x56,0xC8,0xCF,0xAA
.byte 0xFD,0xFE,0xD7,0x6B,0xA6,0xA6,0xB4,0x4C,0x90,0xBB
.byte 0x62,0x6B,0x90,0x81,0x2B,0xEB,0xBB,0x90,0x99,0x10
.byte 0x02,0x15,0x13,0x70,0x56,0xC1,0x27,0x56,0x70,0xE9
.byte 0x86,0x2D,0x18,0x95,0x5E,0x95,0x2D,0x0D,0xE7,0x0A
.byte 0xDE,0xCD,0x27,0x4C,0x05,0xB4,0x9E,0x70,0xD9,0x7A
.byte 0x52,0xFF,0xB1,0xF5,0x10,0xCB,0x6B,0x52,0xF7,0x20
.byte 0x12,0xA2,0xD7,0x56,0x86,0xAE,0xDA,0x24,0xE9,0x1A
.byte 0xAE,0xC1,0xD5,0x7A,0xA2,0x29,0x81,0x90,0x05,0x2F
.byte 0xBB,0x12,0xFB,0x20,0x75,0x35,0x7A,0x3F,0xAA,0x22
.byte 0x8B,0xB8,0x12,0x12,0xBE,0xA2,0xBE,0xFA,0xC1,0x5E
.byte 0x5E,0x01,0xDA,0xFA,0xF6,0x9E,0x70,0x86,0x38,0x56
.byte 0xDC,0xD5,0xCD,0xB1,0x5E,0x2B,0x5A,0xB1,0xDC,0x8B
.byte 0x3D,0xFF,0x24,0x16,0x4F,0xF5,0xE7,0xC1,0x48,0xAE
.byte 0xFF,0x4F,0x3F,0xFC,0x09,0xDA,0x29,0x95,0x16,0x99
.byte 0x31,0x7A,0x0F,0x45,0x62,0x3D,0x42,0x12,0xE4,0x67
.byte 0xA6,0x09,0xD7,0x90,0x7A,0xD7,0xE0,0x86,0xDE,0x75
.byte 0xE6,0x2D,0xFA,0x1D,0x5E,0x5E,0x5E,0x7F,0x1C,0x7F
.byte 0xF6,0x0D,0xC1,0x3D,0xD1,0x75,0xBB,0xB4,0xA2,0xFA
.byte 0x2D,0x0B,0xA6,0xE0,0xAA,0x62,0xFC,0x16,0x0D,0xE9
.byte 0x56,0xCF,0xEE,0x1D,0x0E,0x81,0x7F,0x38,0xDA,0x26
.byte 0xD7,0x24,0x99,0xCF,0xA2,0xD7,0x3D,0x70,0xCB,0x56
.byte 0xC1,0x3F,0x9E,0x0D,0x33,0x75,0x17,0x4C,0x56,0xED
.byte 0xF6,0xF8,0x81,0x7F,0xB1,0x08,0xD5,0x31,0x5E,0x8B
.byte 0x0C,0x31,0x38,0xD9,0x90,0xB4,0x2D,0x99,0xFF,0xAE
.byte 0xD7,0x4C,0x56,0x5E,0x7F,0xCB,0xAE,0x00,0xEC,0x33
.byte 0xB4,0xD1,0x31,0x4F,0x01,0x81,0x09,0x99,0x7F,0x19
.byte 0xEA,0xD7,0x81,0x5E,0xAA,0x5A,0x8B,0x5E,0x04,0xA2
.byte 0x6B,0x4C,0xF7,0x10,0x81,0xF2,0x9E,0x3F,0x5A,0xA6
.byte 0x9E,0x1C,0xF0,0xDE,0xB1,0x24,0xF2,0x56,0x5A,0xC6
.byte 0x8B,0x7F,0xE9,0xA6,0xF6,0xDF,0x90,0xCD,0xF5,0xF2
.byte 0xEE,0x2D,0xC8,0x16,0x52,0x52,0x70,0xEF,0x4C,0x81
.byte 0xF0,0x5E,0x05,0xFB,0xF6,0x7F,0x5A,0xF5,0xC3,0x52
.byte 0xCB,0xC6,0x8B,0x00,0xD9,0x5E,0x70,0x22,0x10,0x56
.byte 0x81,0x45,0xE0,0xDA,0x1C,0xEF,0x52,0x02,0x19,0xAA
.byte 0x67,0x1D,0xC6,0x5A,0xD5,0x4F,0xB4,0xB8,0x81,0x27
.byte 0x13,0x27,0x2F,0xF3,0xEA,0x86,0x13,0x4F,0x99,0x5A
.byte 0xA2,0x4C,0x7A,0x6B,0xDA,0x38,0x45,0x24,0x99,0xC6
.byte 0x26,0x7F,0x31,0x48,0x6B,0xCD,0x8B,0xEB,0xC1,0xD5
.byte 0xF7,0x3F,0x95,0x15,0xD1,0xE2,0x48,0x12,0x17,0xE6
.byte 0x62,0x48,0xAE,0xEE,0xEC,0x9E,0x62,0x62,0x86,0x86
.byte 0xF0,0x27,0x5E,0x04,0x4F,0x33,0x01,0x1A,0x1E,0x1C
.byte 0xF2,0x16,0x2D,0xEC,0xDF,0x2B,0x4F,0xE2,0xAE,0xBE
.byte 0x90,0x56,0x2D,0x13,0xAA,0x02,0x07,0x3F,0xA6,0xDE
.byte 0x81,0xDF,0x2F,0x48,0x5E,0x35,0x8B,0x70,0x01,0x5A
.byte 0x45,0x45,0xEF,0x05,0x35,0xC1,0xA6,0x07,0xFF,0x10
.byte 0x42,0xB4,0x00,0x4F,0x2D,0x8B,0x67,0x31,0x1A,0x09
.byte 0xBE,0x21,0x67,0x3D,0x04,0x3F,0xC8,0x4C,0x86,0xE8
.byte 0x7F,0x81,0x90,0x99,0x70,0x48,0x0E,0xCD,0x7F,0x2F
.byte 0x52,0x26,0x29,0xB4,0x67,0x7F,0x45,0x9E,0x7F,0x62
.byte 0x81,0xA2,0xB8,0x3F,0x62,0x75,0xDC,0xB1,0x67,0x0D
.byte 0xB1,0x48,0x4F,0x31,0xA2,0x4C,0xB1,0xED,0x56,0xB8
.byte 0x81,0xF4,0x18,0xB1,0x70,0x14,0xD3,0xA2,0x75,0xB8
.byte 0x27,0xFB,0x86,0x52,0x0E,0xF6,0xFC,0xF4,0xA2,0x11
.byte 0x09,0x29,0x62,0x2F,0x8B,0x8B,0x4F,0x1C,0xE3,0x62
.byte 0x3F,0x81,0x86,0x8B,0xC8,0x27,0x2D,0xB8,0x31,0xA2
.byte 0xDE,0x26,0x31,0x81,0x5A,0xF0,0x86,0x48,0x99,0x7A
.byte 0x75,0xED,0xE8,0x0F,0x08,0xDC,0x5E,0x35,0x2D,0x14
.byte 0x95,0x5A,0x42,0x4C,0x81,0xAE,0x95,0xDC,0x01,0xD7
.byte 0x70,0x31,0x90,0xAE,0xDC,0xB1,0xCB,0x95,0x42,0x86
.byte 0x0E,0x29,0x56,0xA6,0x9E,0x16,0x8B,0xBB,0x4C,0xE0
.byte 0xDC,0xA2,0x67,0xCD,0x29,0x27,0x62,0x75,0xAA,0xAA
.byte 0x81,0x31,0xAA,0x9E,0x81,0xCD,0x4F,0x67,0x56,0x95
.byte 0x24,0x2B,0x81,0x62,0x62,0x6B,0x07,0x9E,0xAA,0xAA
.byte 0x56,0xDE,0x5A,0xF4,0xFC,0x1E,0x5A,0xAE,0x56,0xF2
.byte 0x81,0x45,0xD1,0x7F,0xCB,0x29,0x8B,0x7A,0xB1,0xD9
.byte 0x99,0x5E,0x56,0x1E,0x3A,0x42,0x6B,0x24,0xB1,0xBB
.byte 0x18,0xF3,0x01,0x2D,0x2B,0xD5,0x62,0x18,0x12,0x7F
.byte 0xE6,0x2F,0xB8,0xB1,0xAA,0x6B,0x99,0x19,0x7F,0xF9
.byte 0xDE,0x2D,0xA6,0xD7,0xD3,0x22,0x86,0x45,0xD9,0xEC
.byte 0x16,0x45,0x67,0x75,0xD7,0x86,0x4C,0x3F,0x67,0xA6
.byte 0xA2,0x09,0xA2,0xB8,0x4C,0x52,0x08,0x99,0xFC,0x8B
.byte 0x6B,0x20,0xD5,0x22,0xF2,0xE0,0xAA,0x62,0x95,0xD3
.byte 0xB4,0x2B,0xCB,0x48,0xE2,0x45,0xA2,0x81,0xEE,0x10
.byte 0x7F,0xDF,0xDE,0xB1,0x67,0x5A,0x2D,0x6B,0x67,0x24
.byte 0x13,0xF7,0xCF,0x45,0x67,0x45,0x75,0x8B,0x81,0x5A
.byte 0xDA,0x70,0xDC,0x3F,0x38,0xE9,0x90,0x0C,0xAA,0x03
.byte 0xB1,0xB1,0xE0,0x22,0x3F,0xD1,0x07,0x2F,0x95,0x75
.byte 0x42,0x26,0x7F,0x8B,0x3D,0x52,0x81,0x45,0x1C,0xDF
.byte 0xC3,0xD9,0xA6,0x09,0xAA,0x19,0x45,0x8B,0x3A,0x12
.byte 0xE8,0xB8,0x1E,0x29,0x67,0xA6,0xAA,0x01,0x1D,0xAA
.byte 0x14,0x90,0xE8,0x90,0x0E,0x48,0xFA,0xF0,0x95,0x05
.byte 0x0F,0x86,0xFE,0x4C,0xCD,0xD1,0xA2,0x1A,0x4F,0xAE
.byte 0xB8,0xC3,0x20,0x67,0xE3,0xDF,0xF5,0xA6,0x0A,0xFE
.byte 0x04,0xD9,0x95,0xF4,0xD9,0x29,0xFE,0x56,0x75,0x81
.byte 0x42,0x75,0x48,0x7F,0xA6,0x6B,0x2B,0x70,0x7F,0x99
.byte 0x01,0xE7,0xA6,0x33,0x7F,0x08,0xAA,0x4F,0xA2,0xD7
.byte 0xF4,0xD9,0x38,0x1D,0x75,0xAE,0x67,0x2B,0x21,0xB1
.byte 0x70,0x7F,0x5E,0x33,0x62,0x70,0x13,0x81,0xEC,0x4F
.byte 0x6B,0xC1,0x62,0x67,0xE0,0xCF,0xCF,0xBB,0x15,0x38
.byte 0x0A,0x3D,0xEF,0x81,0x81,0x70,0x24,0xCB,0xCF,0xF7
.byte 0x99,0xEE,0x3D,0xDE,0x7F,0x9E,0xE8,0x95,0xC1,0xD5
.byte 0x70,0x4F,0x0A,0xB1,0x06,0x2B,0x12,0xF4,0x62,0xAA
.byte 0x20,0x6B,0xA6,0x48,0x81,0xF5,0x90,0xBB,0xD3,0x18
.byte 0x48,0xA2,0x26,0x00,0x7A,0x5E,0x0E,0x2D,0x01,0x6B
.byte 0xA6,0xA2,0xCB,0x3D,0xBB,0x7F,0x99,0x8B,0x4C,0x70
.byte 0x6B,0xB1,0xE0,0xFC,0x5E,0x6B,0x52,0x06,0xDE,0x5E
.byte 0x2F,0x48,0xCB,0x2F,0xB4,0x06,0x90,0x2B,0x81,0xE0
.byte 0xB8,0x3F,0x7A,0xE9,0x24,0xE7,0x5E,0x81,0xA2,0x86
.byte 0xA6,0x27,0x5E,0x99,0x95,0xBB,0xB1,0xCF,0x27,0x99
.byte 0x75,0x02,0x27,0x95,0x8B,0x62,0x90,0xDF,0xAE,0xC6
.byte 0x18,0xB8,0xED,0x4F,0x38,0xCD,0x70,0x99,0x06,0x24
.byte 0x70,0x02,0x70,0x62,0x70,0xEA,0x0E,0xAA,0xE7,0x2D
.byte 0xD1,0x31,0xC3,0xB4,0x15,0x7F,0x03,0x52,0xC1,0xBB
.byte 0x5E,0x9E,0x56,0x56,0x52,0x99,0x2D,0x86,0xC6,0xB4
.byte 0x86,0x99,0xD7,0x95,0xBE,0xD7,0xED,0x81,0xA2,0xED
.byte 0xD3,0x12,0x95,0xAE,0x5E,0x48,0xED,0xA2,0x62,0x1A
.byte 0x86,0xB1,0x12,0x35,0x0B,0xC3,0x6B,0xDC,0x5A,0x24
.byte 0x75,0x5E,0x70,0x5A,0x11,0xCF,0x3F,0x3F,0x9E,0x24
.byte 0x6B,0x7F,0xFC,0xC6,0x0B,0x56,0xFC,0x6B,0x95,0x08
.byte 0x67,0x6B,0xC6,0x7F,0x6B,0xA6,0x75,0x7F,0x48,0xAE
.byte 0xA2,0x00,0x02,0x7A,0x1E,0x07,0x22,0x52,0x48,0xD7
.byte 0xE6,0x8B,0x02,0x95,0xA2,0xB4,0xDC,0xED,0x52,0xA2
.byte 0x2F,0x5E,0xF6,0xD5,0x81,0xB4,0xAE,0x22,0xF5,0x9E
.byte 0x86,0xC1,0x5E,0xBB,0x33,0xDC,0x75,0x1C,0x22,0x35
.byte 0xB4,0x7F,0xDC,0x0C,0x8B,0xF9,0x35,0x4C,0xAE,0xB4
.byte 0x6B,0xF7,0xB1,0x12,0x0B,0x38,0xFC,0x7F,0x1A,0x52
.byte 0x2D,0xB8,0xFC,0x52,0xAE,0xF1,0xC3,0xB8,0xEF,0xA6
.byte 0xB4,0x62,0x90,0xCF,0x7A,0xC8,0x6B,0x99,0xBB,0x18
.byte 0x7A,0xBE,0x56,0xCB,0x11,0x45,0x45,0x90,0x8B,0x01
.byte 0x4F,0xF8,0xD7,0x99,0xDC,0xCF,0x95,0x81,0x2B,0xA6
.byte 0x1A,0x7F,0x5E,0xD7,0x00,0x7F,0xBB,0xC6,0x4C,0xBE
.byte 0xE4,0x90,0x09,0x2D,0x1D,0xA6,0x95,0x81,0x24,0x3D
.byte 0xEA,0xAA,0xBE,0xEF,0x95,0x7F,0xA2,0xB1,0x00,0xFF
.byte 0x0A,0x52,0x2B,0x4F,0xC8,0xBB,0x5A,0xC1,0x4C,0x15
.byte 0xB4,0xBB,0x86,0xB4,0x9E,0x6B,0x7A,0xD9,0x62,0x56
.byte 0xEB,0x8B,0x5E,0xB4,0xFA,0x62,0xBE,0x52,0x5E,0x90
.byte 0x0B,0x52,0x7A,0xF0,0x35,0xE8,0xEA,0x04,0x95,0x81
.byte 0xED,0x3D,0x81,0x2D,0x5A,0x7F,0x45,0x62,0x5A,0x70
.byte 0x18,0xB4,0x3A,0x67,0xAE,0x6B,0x6B,0xCD,0x95,0xD9
.byte 0x7A,0x4C,0x62,0xF6,0x7A,0x5E,0xFE,0xC6,0xF5,0x19
.byte 0x6B,0xD1,0x38,0x35,0x67,0x7F,0x5A,0x15,0x52,0xA6
.byte 0xB4,0x16,0x10,0x19,0xD3,0xFB,0xAE,0x81,0x4F,0xB1
.byte 0xD9,0x0D,0x2B,0x17,0x45,0x67,0xE2,0x1A,0xB8,0xED
.byte 0x18,0xF8,0xA2,0x1D,0xC6,0x7F,0xAA,0x7A,0xAE,0xA2
.byte 0x8B,0xEF,0x14,0x10,0xDE,0x35,0xB1,0x5E,0xE2,0xAA
.byte 0x67,0xAA,0x01,0x2F,0x24,0x81,0xE8,0x67,0x19,0x7F
.byte 0x75,0x3D,0xDE,0xDF,0x02,0xD7,0x08,0xC3,0xB8,0x95
.byte 0xBE,0x6B,0xFF,0x67,0xA2,0x14,0xEC,0xB8,0x5E,0x24
.byte 0x12,0x2B,0x81,0x9E,0x04,0xC6,0x6B,0x5E,0xA6,0x9E
.byte 0x7F,0x48,0x52,0x90,0x3F,0x27,0xBE,0x5A,0xCF,0xFE
.byte 0x3F,0x67,0x8B,0x7F,0x6B,0x22,0x02,0x5E,0xD9,0x81
.byte 0xF2,0x2F,0x5E,0xFC,0x16,0x52,0x0C,0x6B,0xCB,0xBB
.byte 0xD1,0x4F,0x5E,0x67,0xE0,0xF5,0x5E,0x0B,0xE0,0x4F
.byte 0x45,0x67,0x2F,0xAA,0xBE,0x81,0x24,0x26,0x5E,0x5A
.byte 0x45,0x5E,0x22,0x52,0xEB,0xB8,0x38,0xEB,0xBB,0x8B
.byte 0x2D,0x5E,0x70,0x09,0x1D,0x99,0x86,0xA2,0x09,0x99
.byte 0x7F,0x18,0x3F,0x81,0x18,0x7F,0x31,0xE2,0x09,0x62
.byte 0xBE,0x56,0x09,0xBB,0x75,0x9E,0xB4,0x5E,0x24,0x99
.byte 0x62,0xAE,0x7A,0x81,0x99,0xCF,0x1C,0x26,0x05,0xCB
.byte 0x22,0x95,0x26,0x62,0x7F,0x56,0xA6,0x31,0xAA,0x08
.byte 0xCB,0xC1,0xFC,0x45,0xEF,0xA6,0xD7,0x0F,0xDC,0x5E
.byte 0xEB,0x52,0xCB,0x45,0xB1,0x12,0xF7,0x75,0xE0,0x9E
.byte 0x6B,0x56,0x04,0x45,0xF4,0xB1,0x7A,0x99,0x29,0xE7
.byte 0xB8,0xD7,0x9E,0x1D,0x67,0x04,0x86,0x90,0x5E,0x31
.byte 0xB8,0x7F,0x86,0xE3,0xC3,0x67,0xC3,0x8B,0x10,0x9E
.byte 0xB8,0xE6,0x45,0x2B,0x09,0x7A,0xF5,0xB1,0xB8,0xA2
.byte 0x2F,0x11,0x62,0x48,0x1D,0xBE,0x56,0x75,0x24,0x95
.byte 0xDE,0x95,0x81,0x56,0x75,0x09,0x67,0x19,0x45,0x7F
.byte 0x7F,0xD3,0x20,0x01,0x48,0xC8,0x99,0xF2,0x70,0xAA
.byte 0x4F,0x99,0x7F,0x7F,0xBE,0x20,0xB8,0xC8,0xB1,0x86
.byte 0x11,0x70,0x13,0xC1,0xA6,0xC6,0x90,0x7A,0x81,0x11
.byte 0xB1,0x81,0x04,0x10,0x9E,0x01,0x7A,0x4C,0x3F,0x8B
.byte 0xA6,0xD5,0x99,0xF7,0xF0,0x52,0x13,0x81,0x86,0xD9
.byte 0x75,0x90,0x8B,0xA2,0xDE,0xED,0xD3,0xCB,0x06,0xB1
.byte 0x42,0x4F,0x2D,0x1C,0x7F,0xCB,0x7F,0x3D,0x22,0x45
.byte 0x5E,0x67,0xCB,0x3A,0xBB,0xC6,0x3A,0xE2,0xC8,0xF7
.byte 0x90,0x8B,0xB1,0x7F,0x45,0x3F,0x56,0x75,0x95,0x45
.byte 0x4F,0x42,0xDF,0x1E,0x95,0x86,0x2D,0x86,0x35,0xD1
.byte 0xF8,0x45,0x01,0xC3,0x7F,0x22,0x3F,0x16,0x3F,0x33
.byte 0xFB,0x5E,0x11,0x90,0x0A,0x86,0xC8,0xDC,0x13,0x17
.byte 0x7A,0xBB,0x86,0xA6,0x7A,0x95,0x81,0x26,0xC3,0x38
.byte 0x15,0x95,0xB1,0x7A,0x02,0xAA

.ascii	"MEMY"
Memory:
.ALIGN
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x4C60144C,0x7B4C605A,0x60062060,0x198D01A9,0xEA314CD0,0x00A953A0
.word 0x88033399,0x00A9FAD0,0xAC03778D,0x8C8867F4,0xEFAC037C,0x03858C67,0x20647520,0x1CA064A0,0x009900A9,0x02A799D4
.word 0xADF71088,0x188D67FA,0x67F9ADD4,0xA9D4178D,0xD4048D08,0x8DD40B8D,0xA960D412,0x03148D09,0x158D60A9,0x8D64A903
.word 0x1BA9D012,0xA9D0118D,0xD01A8D81,0x0E8D00A9,0xEE6058DC,0x7CAD037C,0x67F4CD03,0x00A905D0,0xAD037C8D,0x05D065F0
.word 0x7A2000A2,0x65F1AD62,0x07A205D0,0xAD627A20,0x05D065F2,0x7A200EA2,0x037CAD62,0x3D2003D0,0x0380AD64,0x033013F0
.word 0xAD0380CE,0x168D02BD,0x846D18D4,0x02BD8D03,0x67F8AD60,0xF8F00129,0xED3800A9,0x848D0384,0x67FCAD03,0x4C03808D
.word 0x402060BB,0x0337BD61,0x39BC39F0,0x6900B903,0x9D03633D,0x00B9D404,0xD0082969,0x6800B923,0x05D07FC9,0xD067FFAC
.word 0x357D1808,0x03359D03,0x6527B9A8,0xB9D4019D,0x009D6587,0x612E4CD4,0xBD612E20,0x03D0034E,0x4C62144C,0x39FE6187
.word 0x0339BD03,0xD00337DD,0x0338BD03,0x6003399D,0x9D02A9BD,0xAABDD402,0xD4039D02,0xD00373BD,0x02A9BD1B,0x03347D18
.word 0xBD02A99D,0x006902AA,0xCD02AA9D,0x03D067FD,0x600373FE,0x3802A9BD,0x9D0334FD,0xAABD02A9,0x9D00E902,0xFECD02AA
.word 0xDE03D067,0xBD600373,0x01D0034E,0xF002C960,0x5EBD183B,0x02A77D03,0xBD02A79D,0xA87D034F,0x02A89D02,0xDD02A8BD
.word 0x04F0034D,0x08B05990,0xDD02A7BD,0x4F90034C,0x9D034CBD,0x4DBD02A7,0x02A89D03,0x4E9D00A9,0x62074C03,0x02A7BD38
.word 0x9D035EFD,0xA8BD02A7,0x034FFD02,0xBD02A89D,0xA8DD034D,0xF0209002,0xBD08B002,0xA7DD034C,0xBD149002,0xA79D034C
.word 0x034DBD02,0xA902A89D,0x034E9D00,0xBD62074C,0x009D02A7,0x02A8BDD4,0x60D4019D,0xF0033ABD,0x034BBD5E,0x4BDE06F0
.word 0x62774C03,0xBD0349FE,0xF6CD0349,0xBD0DD067,0x0149034A,0xA9034A9D,0x03499D00,0xD0034ABD,0xA7BD1815,0x033A7D02
.word 0xBD02A79D,0x006902A8,0x4C02A89D,0xBD386268,0x3AFD02A7,0x02A79D03,0xE902A8BD,0x02A89D00,0x033ABD18,0xDD03747D
.word 0x0390033A,0x4C033A9D,0x7D8E6207,0x037CAD03,0x8B2003D0,0x037DAE62,0xA960E24C,0x03758D00,0x850360BD,0x0361BDA6
.word 0x77ACA785,0xF0A6B103,0xC90B1071,0xEE06F0FF,0xAE4C0375,0xA2C86062,0x10A6B100,0x3F29E801,0x7DAE8A48,0x034E9D03
.word 0x5F7D1868,0x90B9A803,0xB9A58566,0x379D6660,0x6600B903,0xB903399D,0x389D6630,0xADA5A403,0x17D00375,0xD067FBEC
.word 0x67A0B912,0xB902BD8D,0x848D67C0,0x67FCAD03,0xAC03808D,0xB1C80377,0xD04029A6,0xA6B18811,0x7D187F29,0x6D180364
.word 0x4CAA0386,0xB188631C,0xAA7F29A6,0x03D000E0,0xAC64204C,0x27BD037D,0x03788D65,0x8D6587BD,0x998A0379,0x7DAE0335
.word 0x034EB903,0x79AD28F0,0x034C9D03,0x9D0378AD,0xA8BD034D,0x034DDD02,0x02F01190,0xA7BD08B0,0x034CDD02,0x02A90590
.word 0x4C034E9D,0x78AD6378,0x02A89D03,0xADD4019D,0xA79D0379,0xD4009D02,0x75ADA5A4,0x4C03F003,0xC0B96403,0xD4059D66
.word 0x9D66E0B9,0x00A9D406,0xB903739D,0x4A4A6700,0xAA9D4A4A,0x6700B902,0x0A0A0A0A,0xB902A99D,0x349D6720,0x6740B903
.word 0xF0033A9D,0xA9073014,0x03749D00,0x7F290BF0,0xAD03749D,0x3A9D67F5,0x9D00A903,0xF7AD034A,0x03499D67,0x4A6760B9
.word 0x4B9D4A4A,0x6760B903,0x4F9D0729,0x6780B903,0xBD035E9D,0x1FD00337,0x9D0338BD,0x049D0362,0x0339BDD4,0x9DD4049D
.word 0xA96002AB,0x034A9D00,0x9D67F7AD,0xBC600349,0xF7A90339,0xB903639D,0xF6296900,0x60D4049D,0xBD037DAE,0x0FD00337
.word 0xDD0362BD,0x06F002AB,0x9DD4049D,0xA96002AB,0x03639DF6,0x0377EE60,0xAD0377EE,0xF1CD0377,0xA929D067,0x03778D00
.word 0xAC0376EE,0xCCC867F3,0x15D00376,0xAC0385EE,0xCCC867F0,0x06D00385,0x8C67EFAC,0x75200385,0x64A02064,0x0385AC60
.word 0x8D7000B9,0xF28D0376,0x7050B967,0xB967F38D,0x868D70A0,0x70F0B903,0xB9035F8D,0x668D7140,0x7190B903,0x60036D8D
.word 0xB90376AC,0x648D6D00,0x6A00B903,0x8C64E020,0x608E0361,0x0376AC03,0x8D6E00B9,0x00B9036B,0x64E0206B,0x8E03688C
.word 0x76AC0367,0x6F00B903,0xB903728D,0xE0206C00,0x036F8C64,0x60036E8E,0xF1ADA685,0xA2A78567,0x4600A908,0x180390A7
.word 0x6E6AA665,0xD0CA037E,0x037F8DF2,0x037EAD18,0x7E8D0069,0x037FAD03,0xAEA87269,0x5360037E,0x444E554F,0x5453414D
.word 0x42205245,0x4F532059,0x53454445,0x0154464F,0x01010101,0x01010101,0x02020101,0x02020202,0x03030302,0x04040303
.word 0x05050404,0x06060605,0x08080707,0x0A0A0909,0x0D0D0C0B,0x11100F0E,0x15141312,0x1B1A1817,0x22201F1D,0x2B292724
.word 0x3734312E,0x45413E3A,0x57524E49,0x6E68625C,0x8B837C75,0xAFA59C93,0xDDD0C4B9,0x1700F8EA,0x5F4B3927,0xBAA18A74
.word 0x2D0EF0D4,0xBE96714E,0x744314E8,0x5A1CE1A9,0x7C2DE29C,0xE88528CF,0xB437C152,0xF75AC539,0xD10A4F9E,0x686E82A3
.word 0xEEB38A71,0xA2159E3C,0xD0DC0446,0xDD6714E2,0x44293C79,0xA1B8088D,0xBACD28C5,0x875378F1,0x4271101A,0x749B4F89
.word 0x0EA6F0E2,0x00002033,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x0D000700,0x1B14120D
.word 0x3C312622,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x10060C06,0x1C151310,0x3D372C25,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x12070D07,0x221B1412,0x433C3126,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00020100,0x04040302,0x04060605,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x0000000E,0x00000900,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0xE155A3E8,0x00A529B9,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x60800080,0x00008018,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x3000F000,0x00000403,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x09000000,0x00005000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x40000000,0x00004000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x60000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x50000000,0x000000A8,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0xFD000000,0x000000FA,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x0D004000,0x01040904,0x071FF200,0x4E010E20,0xFC227F00,0x00FFFEFD,0x08F0307F,0x227F0000
.word 0x0000EC14,0x05000300,0x0000F800,0x00040005,0x000000F7,0x7F000000,0xF903E024,0x00030004,0x237F0007,0x03F904E0
.word 0xF9000400,0x05000400,0x0000F700,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x11418109,0x09111111,0x80808181,0x41810980,0x49091181,0x41414109,0x09414141,0x41414141,0x41094141,0x81091941
.word 0x21212181,0x21212121,0x81810921,0x21212121,0x21212121,0x41414109,0x00414141,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x01010101,0x01010108,0x01010108,0x00000108,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x02020000,0x02020704
.word 0x02020704,0x00000704,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x03030000,0x03030605,0x03030605,0x00000605,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x0000FB00,0x0000FB00,0x0000FB00
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x0000000D,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00004008,0x0000413C,0x00000938,0x00004208,0x00004308,0x00000938,0x00004008,0x00004408
.word 0x00000938,0x00004408,0x00004008,0x00000938,0x00004308,0x00004008,0x00004008,0x00004130,0x00FF0514,0x00FF00FF
.word 0x000000FF,0x00FF0514,0x000000FF,0x00000514,0x00FF050F,0x00FF0511,0x00FF0514,0x00FF00FF,0x00000514,0x00FF0520
.word 0x00000000,0x00000514,0x00FF051B,0x00FF051D,0x00FF083F,0x00FF00FF,0x00FF08BF,0x00FF083D,0x00FF00FF,0x00FF08BF
.word 0x00FF083C,0x00000000,0x00000630,0x00000630,0x00000630,0x00000630,0x00000733,0x00000000,0x00000630,0x00000000
.word 0x00FF0511,0x00FF00FF,0x000000FF,0x00FF0511,0x000000FF,0x00000511,0x00FF0511,0x000000FF,0x00FF0511,0x000000FF
.word 0xA09D051A,0x00000000,0xA09B0519,0x00000000,0x00FF0518,0x00FF0516,0x00FF0844,0x00FF00FF,0x00FF08C4,0x00FF0843
.word 0x00FF00FF,0x00FF08C3,0x00FF083C,0x00FF00FF,0x00FF08BC,0x00FF08BC,0x00FF00FF,0x000008BC,0x00FF0B38,0x00000000
.word 0x00FF0B38,0x00000000,0x00FF0843,0x00FF00FF,0x00FF08C3,0x00FF083F,0x00FF00FF,0x00FF08BF,0x00FF083A,0x00FF00FF
.word 0x000008BA,0x000008BA,0x00000000,0x000008BA,0x000008BA,0x00000000,0x000008BA,0x000008BA,0x00FF050F,0x00FF00FF
.word 0x000000FF,0x00FF051B,0x00FF072E,0x00000000,0x00FF050F,0x00FF072E,0x000000FF,0x00FF050F,0x00FF072E,0x000000FF
.word 0x00FF072E,0x000000FF,0x00FF0513,0x000000FF,0x00004008,0x0000413C,0x00000A35,0x00004208,0x00004308,0x00000A35
.word 0x00004008,0x00004408,0x00000A35,0x00004408,0x00004008,0x00000A35,0x00004308,0x00004008,0x00004008,0x00004130
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000
.word 0x00000000,0x00000000,0x00000000,0x00000000

@--------------------------------------
@--# C64'er Emulator Code V0.3 Beta #--
@--# no bugs known yet!             #--
@--# (w) 1995 Denys Bogatz, (c) DDS #--
@--#     (for Acorn Computers)      #--
@--#    modified in 2002 for GBA    #--
@--------------------------------------

@----------------------------------------------------------------------------

AK:	.word	0x00000000			@ akku
XR:	.word	0x00000000			@ xregister
YR:	.word	0x00000000			@ etc...
LA:	.word	0x00000000
FL:	.word	0x00000000

@----------------------------------------------------------------------------
.ascii	"MOS6502-Emulator"
.ALIGN 4
.GLOBAL C64Init

C64Init:
	@ -------------------------------------------------------------------
	@ on entry:
	@ R0 = Akku	(normally be used to select sound!)
	@ R1 = PC	(where to jump (c64-subroutine))
	@ -------------------------------------------------------------------
	STMFD	R13!,{R0-R12,R14}	@ preserve registers
	MOV	R4,R0			@ preserve AKKU
	MOV	r3,r1			@ get jumpaddress into r3!
	
	CMP	R3,#0
	BEQ	ExitCBM
	ADRL	R0,Memory		@ base of C64-Memory
	
	MOV	R5,#0
	STR	R5,[R0,#0x0318]		@ NMI INIT
	ADD	R1,R0,#0x10000
	STR	R5,[R1,#-8]		@ also NMI INIT
	
	ADD	R0,R0,#0xD400		@ base of SID-Registers
	MOV	R1,#0
	MOV	R2,#0x13
InitSid:ldr	R2,[R0,R1]
	ADD	R1,R1,#4
	CMP	R1,#0x0100
	BLT	InitSid
	MOV	R1,#15
	STRB	R1,[R0,#24]

	MOV	R0,R4			@ AKKU
	MOV	R1,R4			@ XReg <- AKKU
	MOV	R2,R4			@ YReg <- AKKU
	@	R3			@ C64-Program-Counter
CBMPlay:MOV	R4,#0xFF		@ FLAGS
	MOV	R5,#0xF8		@ STACKPOINTER
	
	ADRL	R11,Memory		@ where C64-Memory is located in GBA
	ADD	R3,R3,R11		@ add jump-address to it (where C64-code is to be executed)
	
	MOV	R6,#0			@ just to be safe add return-address to c64-stack
	ADD	R8,R11,#0x0100		@ (when something went wrong a jmp$0 is executed)
	STRB	R6,[R8,R5]
	SUB	R5,R5,#1
	STRB	R6,[R8,R5]
	SUB	R5,R5,#1
	MOV	R6,#0x80		@ add "emulation-break"-opcode to memory located at $0
	STRB	R6,[R11]		@ ...so jmp$0 works (see above)
	ADRL	R10,ADSR_64

	CMP	R3,R11
	BEQ	ExitCBM


	@ -------------------------------------------------------------------
	@ This is the final emulation-loop! (most opcodes jump back to this!)
	@ -------------------------------------------------------------------
CBMLoop:LDRB	R6,[R3]			@ get next opcode
	ADD	PC,PC,R6,LSL#2		@ do something like "on a goto"
	MOV	R0,R0			@ need this NOP 'coz of ARM-pipeline!
	B	Q00			
	B	Q01
	B	Q02
	B	Q03
	B	Q04
	B	Q05
	B	Q06
	B	Q07
	B	Q08
	B	Q09
	B	Q0A
	B	Q0B
	B	Q0C
	B	Q0D
	B	Q0E
	B	Q0F
	B	Q10
	B	Q11
	B	Q12
	B	Q13
	B	Q14
	B	Q15
	B	Q16
	B	Q17
	B	Q18
	B	Q19
	B	Q1A
	B	Q1B
	B	Q1C
	B	Q1D
	B	Q1E
	B	Q1F
	B	Q20
	B	Q21
	B	Q22
	B	Q23
	B	Q24
	B	Q25
	B	Q26
	B	Q27
	B	Q28
	B	Q29
	B	Q2A
	B	Q2B
	B	Q2C
	B	Q2D
	B	Q2E
	B	Q2F
	B	Q30
	B	Q31
	B	Q32
	B	Q33
	B	Q34
	B	Q35
	B	Q36
	B	Q37
	B	Q38
	B	Q39
	B	Q3A
	B	Q3B
	B	Q3C
	B	Q3D
	B	Q3E
	B	Q3F
	B	Q40
	B	Q41
	B	Q42
	B	Q43
	B	Q44
	B	Q45
	B	Q46
	B	Q47
	B	Q48
	B	Q49
	B	Q4A
	B	Q4B
	B	Q4C
	B	Q4D
	B	Q4E
	B	Q4F
	B	Q50
	B	Q51
	B	Q52
	B	Q53
	B	Q54
	B	Q55
	B	Q56
	B	Q57
	B	Q58
	B	Q59
	B	Q5A
	B	Q5B
	B	Q5C
	B	Q5D
	B	Q5E
	B	Q5F
	B	Q60
	B	Q61
	B	Q62
	B	Q63
	B	Q64
	B	Q65
	B	Q66
	B	Q67
	B	Q68
	B	Q69
	B	Q6A
	B	Q6B
	B	Q6C
	B	Q6D
	B	Q6E
	B	Q6F
	B	Q70
	B	Q71
	B	Q72
	B	Q73
	B	Q74
	B	Q75
	B	Q76
	B	Q77
	B	Q78
	B	Q79
	B	Q7A
	B	Q7B
	B	Q7C
	B	Q7D
	B	Q7E
	B	Q7F
	B	Q80
	B	Q81
	B	Q82
	B	Q83
	B	Q84
	B	Q85
	B	Q86
	B	Q87
	B	Q88
	B	Q89
	B	Q8A
	B	Q8B
	B	Q8C
	B	Q8D
	B	Q8E
	B	Q8F
	B	Q90
	B	Q91
	B	Q92
	B	Q93
	B	Q94
	B	Q95
	B	Q96
	B	Q97
	B	Q98
	B	Q99
	B	Q9A
	B	Q9B
	B	Q9C
	B	Q9D
	B	Q9E
	B	Q9F
	B	QA0
	B	QA1
	B	QA2
	B	QA3
	B	QA4
	B	QA5
	B	QA6
	B	QA7
	B	QA8
	B	QA9
	B	QAA
	B	QAB
	B	QAC
	B	QAD
	B	QAE
	B	QAF
	B	QB0
	B	QB1
	B	QB2
	B	QB3
	B	QB4
	B	QB5
	B	QB6
	B	QB7
	B	QB8
	B	QB9
	B	QBA
	B	QBB
	B	QBC
	B	QBD
	B	QBE
	B	QBF
	B	QC0
	B	QC1
	B	QC2
	B	QC3
	B	QC4
	B	QC5
	B	QC6
	B	QC7
	B	QC8
	B	QC9
	B	QCA
	B	QCB
	B	QCC
	B	QCD
	B	QCE
	B	QCF
	B	QD0
	B	QD1
	B	QD2
	B	QD3
	B	QD4
	B	QD5
	B	QD6
	B	QD7
	B	QD8
	B	QD9
	B	QDA
	B	QDB
	B	QDC
	B	QDD
	B	QDE
	B	QDF
	B	QE0
	B	QE1
	B	QE2
	B	QE3
	B	QE4
	B	QE5
	B	QE6
	B	QE7
	B	QE8
	B	QE9
	B	QEA
	B	QEB
	B	QEC
	B	QED
	B	QEE
	B	QEF
	B	QF0
	B	QF1
	B	QF2
	B	QF3
	B	QF4
	B	QF5
	B	QF6
	B	QF7
	B	QF8
	B	QF9
	B	QFA
	B	QFB
	B	QFC
	B	QFD
	B	QFE
	B	QFF

@----------------------------------------------------------------------------

IlOpCod:
ExitCBM:STR	R0,AK
	STR	R1,XR
	STR	R2,YR
	STR	R3,LA
	STR	R4,FL
	LDMFD	R13!,{R0-R12,r14}
	bx	lr

@----------------------------------------------------------------------------
Q02:
Q03:
Q04:
Q07:
Q0B:
Q0C:
Q0F:
Q12:
Q13:
Q14:
Q17:
Q1A:
Q1B:
Q1C:
Q1F:
Q22:
Q23:
Q27:
Q2B:
Q2F:
Q32:
Q33:
Q34:
Q37:
Q3A:
Q3B:
Q3C:
Q3F:
Q42:
Q43:
Q44:
Q47:
Q4B:
Q4F:
Q52:
Q53:
Q54:
Q57:
Q5A:
Q5B:
Q5C:
Q5F:
Q62:
Q63:
Q64:
Q67:
Q6B:
Q6F:
Q72:
Q73:
Q74:
Q77:
Q7A:
Q7B:
Q7C:
Q7F:
Q82:
Q83:
Q87:
Q89:
Q8B:
Q8F:
Q92:
Q93:
Q97:
Q9B:
Q9C:
Q9E:
Q9F:
QA3:
QA7:
QAB:
QAF:
QB2:
QB3:
QB7:
QBB:
QBF:
QC2:
QC3:
QC7:
QCB:
QCF:
QD2:
QD3:
QD4:
QD7:
QDA:
QDB:
QDC:
QDF:
QE2:
QE3:
QE7:
QEB:
QEF:
QF2:
QF3:
QF4:
QF7:
QFA:
QFB:
QFC:
QFF:
        B       IlOpCod

@----------------------------------------------------------------------------

Q80:    B       ExitCBM         @ Exit CBM 64 Simulation !!!

@----------------------------------------------------------------------------

Q00:                            @ BRK (NUR MIT NMI SINNVOLL!)
        B       ExitCBM
@ f*ck gcc-asm - don't know how to define binaries...  
@ so here're some defines i need!
.equ	b_00000001,	0x0001
.equ  	b_00000010,	0x0002
.equ	b_00000100,	0x0004
.equ	b_00001000,	0x0008
.equ	b_01000000,	0x0040
.equ  	b_10000000,	0x0080

.equ	b_00111101,	0x003d
.equ	b_01111100,	0x007C
.equ  	b_01111101,	0x007D
.equ	b_11000000,	0x00c0
.equ	b_11111110,	0x00FE
@----------------------------------

Q01:                            @ ORA(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R7,R11]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q05:                            @ ORA ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q06:                            @ ASL ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        MOV     R7,R7,LSL#1
        STRB    R7,[R11,R6]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R7,LSR#8
        ANDS    R7,R7,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q08:                            @ PHP
        ADD     R3,R3,#1
        ADD     R8,R11,#0x100
        STRB    R4,[R8,R5]
        SUB     R5,R5,#1
        B       CBMLoop

@----------------------------------

Q09:                            @ ORA#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q0A:                            @ ASL A
        ADD     R3,R3,#1
        MOV     R0,R0,LSL#1
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q0D:                            @ ORA AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q0E:                            @ ASL AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        MOV     R6,R6,LSL#1
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R6,LSR#8
        ANDS    R6,R6,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R6,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q10:                            @ BPL
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_10000000
        BNE     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop

@----------------------------------

Q11:                            @ ORA(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q15:                            @ ORA ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R6,[R11,R6]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q16:                            @ ASL ZPX
        LDRB    R7,[R3,#1]
        ADD     R3,R3,#2
        ADD     R7,R7,R1
        AND     R7,R7,#0xFF
        LDRB    R6,[R11,R7]
        MOV     R6,R6,LSL#1
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R6,LSR#8
        ANDS    R6,R6,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R6,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q18:                            @ CLC
        ADD     R3,R3,#1
        AND     R4,R4,#b_11111110
        B       CBMLoop

@----------------------------------

Q19:                            @ ORA ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q1D:                            @ ORA ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        ORRS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q1E:                            @ ASL ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        MOV     R6,R6,LSL#1
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R6,LSR#8
        ANDS    R6,R6,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R6,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q20:                            @ JSR
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R7,R7,R8,LSL#8
        ADD     R3,R3,#2
        SUB     R3,R3,R11
        ADD     R8,R11,#0x100
        MOV     R6,R3,LSR#8
        STRB    R6,[R8,R5]
        SUB     R5,R5,#1
        STRB    R3,[R8,R5]
        SUB     R5,R5,#1
        ADD     R3,R7,R11
        B       CBMLoop

@----------------------------------

Q21:                            @ AND(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q24:                            @ BIT ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        AND     R4,R4,#b_00111101
        AND     R7,R6,#b_11000000
        ORR     R4,R4,R7
        ANDS    R6,R0,R6
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop

@----------------------------------

Q25:                            @ AND ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q26:                            @ ROL ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        MOV     R7,R7,LSL#1
        AND     R8,R4,#1
        ORR     R7,R7,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R7,LSR#8
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q28:                            @ PLP
        ADD     R3,R3,#1
        ADD     R8,R11,#0x100
        ADD     R5,R5,#1
        LDRB    R4,[R8,R5]
        B       CBMLoop

@----------------------------------

Q29:                            @ AND#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q2A:                            @ ROL A
        ADD     R3,R3,#1
        MOV     R0,R0,LSL#1
        AND     R7,R4,#1
        ORR     R0,R0,R7
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q2C:                            @ BIT AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        AND     R4,R4,#b_00111101
        AND     R7,R6,#b_11000000
        ORR     R4,R4,R7
        ANDS    R6,R0,R6
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop

@----------------------------------

Q2D:                            @ AND AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q2E:                            @ ROL AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        MOV     R6,R6,LSL#1
        AND     R8,R4,#1
        AND     R4,R4,#b_01111100
        ORR     R6,R6,R8
        ORR     R4,R4,R6,LSR#8
        STRB    R6,[R11,R7]
        ANDS    R6,R6,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R6,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q30:                            @ BMI
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_10000000
        BEQ     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop

@----------------------------------

Q31:                            @ AND(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q35:                            @ AND ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R6,[R11,R6]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q36:                            @ ROL ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        MOV     R7,R7,LSL#1
        AND     R8,R4,#1
        ORR     R7,R7,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R7,LSR#8
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q38:                            @ SEC
        ADD     R3,R3,#1
        ORR     R4,R4,#1
        B       CBMLoop

@----------------------------------

Q39:                            @ AND ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q3D:                            @ AND ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        ANDS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q3E:                            @ ROL ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        MOV     R6,R6,LSL#1
        AND     R8,R4,#1
        ORR     R6,R6,R8
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R6,LSR#8
        ANDS    R6,R6,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R6,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q40:                            @ RTI
        B       ExitCBM

@----------------------------------

Q41:                            @ EOR(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q45:                            @ EOR ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q46:                            @ LSR ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        AND     R8,R7,#1
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R8
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop

@----------------------------------

Q48:                            @ PHA
        ADD     R3,R3,#1
        ADD     R8,R11,#0x100
        STRB    R0,[R8,R5]
        SUB     R5,R5,#1
        B       CBMLoop

@----------------------------------

Q49:                            @ EOR#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q4A:                            @ LSR A
        ADD     R3,R3,#1
        AND     R8,R0,#1
        MOVS    R0,R0,LSR#1
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R8
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop

@----------------------------------

Q4C:                            @ JMP
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R7,R7,R8,LSL#8
        ADD     R3,R7,R11
        B       CBMLoop
        
@----------------------------------

Q4D:                            @ EOR AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q4E:                            @ LSR AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        AND     R8,R6,#1
        MOVS    R6,R6,LSR#1
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R8
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop
        
@----------------------------------

Q50:                            @ BVC
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_01000000
        BNE     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop
        
@----------------------------------

Q51:                            @ EOR(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q55:                            @ EOR ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R6,[R11,R6]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q56:                            @ LSR ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        AND     R8,R7,#1
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R8
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop
        
@----------------------------------

Q58:                            @ CLI
        ADD     R3,R3,#1
        BIC     R4,R4,#b_00000100
        B       CBMLoop

@----------------------------------

Q59:                            @ EOR ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q5D:                            @ EOR ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        EORS    R0,R0,R6
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q5E:                            @ LSR ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        AND     R8,R6,#1
        MOVS    R6,R6,LSR#1
        STRB    R6,[R11,R7]
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R8
        ORREQ   R4,R4,#b_00000010
        B       CBMLoop
        
@----------------------------------

Q60:                            @ RTS
        ADD     R8,R11,#0x100
        ADD     R5,R5,#1
        LDRB    R7,[R8,R5]
        ADD     R5,R5,#1
        LDRB    R8,[R8,R5]
        ADD     R7,R7,R8,LSL#8
        ADD     R3,R7,#1
        ADD     R3,R3,R11
        CMP     R5,#0xF8
        BGE     ExitCBM
        B       CBMLoop
        
@----------------------------------

Q61:                            @ ADC(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q65:                            @ ADC ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q66:                            @ ROR ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        AND     R8,R4,#1
        ORR     R7,R7,R8,LSL#8
        AND     R4,R4,#b_01111100
        AND     R8,R7,#1
        ORR     R4,R4,R8
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q68:                            @ PLA
        ADD     R3,R3,#1
        ADD     R8,R11,#0x100
        ADD     R5,R5,#1
        LDRB    R0,[R8,R5]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q69:                            @ ADC#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q6A:                            @ ROR A
        ADD     R3,R3,#1
        AND     R8,R4,#1
        ORR     R0,R0,R8,LSL#8
        AND     R4,R4,#b_01111100
        AND     R8,R0,#1
        ORR     R4,R4,R8
        MOVS    R0,R0,LSR#1
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q6C:                            @ JMP ($X)
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R6,R7,R8,LSL#8
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        ADD     R3,R7,R11
        B       CBMLoop
        
@----------------------------------

Q6D:                            @ ADC AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q6E:                            @ ROR AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        LDRB    R7,[R11,R6]
        AND     R8,R4,#1
        ORR     R7,R7,R8,LSL#8
        AND     R4,R4,#b_01111100
        AND     R8,R7,#1
        ORR     R4,R4,R8
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q70:                            @ BVS
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_01000000
        BEQ     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop

@----------------------------------

Q71:                            @ ADC(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

Q75:                            @ ADC ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R6,[R11,R6]
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q76:                            @ ROR ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        AND     R8,R4,#1
        ORR     R7,R7,R8,LSL#8
        AND     R4,R4,#b_01111100
        AND     R8,R7,#1
        ORR     R4,R4,R8
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q78:                            @ SEI
        ADD     R3,R3,#1
        ORR     R4,R4,#b_00000100
        B       CBMLoop

@----------------------------------

Q79:                            @ ADC ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q7D:                            @ ADC ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        ADD     R0,R0,R6
        AND     R8,R4,#1
        ADD     R0,R0,R8
        AND     R4,R4,#b_01111100
        ORR     R4,R4,R0,LSR#8
        ANDS    R0,R0,#0xFF
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q7E:                            @ ROR ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R6,R7,R1
        LDRB    R7,[R11,R6]
        AND     R8,R4,#1
        ORR     R7,R7,R8,LSL#8
        AND     R4,R4,#b_01111100
        AND     R8,R7,#1
        ORR     R4,R4,R8
        MOVS    R7,R7,LSR#1
        STRB    R7,[R11,R6]
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q81:                            @ STA(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R6,R7,R8,LSL#8
        STRB    R0,[R11,R6]

        @ ----	check if soundreg (ADSR) got modified (dirty hack coz some players will modify this reg too fast...)
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R0,#1		@ bit0 cleared somewhere in sid (may affect ADSR)
        BNE     CBMLoop
        
        AND     R7,R6,#0xFF
        CMP     R7,#4		@ attack-flag changed! store a value (greater than 2) in our ADSR_64-tab to reflect the change!
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop

@----------------------------------

Q84:                            @ STY ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        STRB    R2,[R11,R6]
        B       CBMLoop

@----------------------------------

Q85:                            @ STA ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        STRB    R0,[R11,R6]
        B       CBMLoop

@----------------------------------

Q86:                            @ STX ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        STRB    R1,[R11,R6]
        B       CBMLoop

@----------------------------------

Q88:                            @ DEY
        ADD     R3,R3,#1
        SUB     R2,R2,#1
        ANDS    R2,R2,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q8A:                            @ TXA
        ADD     R3,R3,#1
        MOVS    R0,R1
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q8C:                            @ STY AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        STRB    R2,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R2,#1
        BNE     CBMLoopTST
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop

@----------------------------------

Q8D:                            @ STA AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        STRB    R0,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R0,#1
        BNE     CBMLoopTST
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
CBMLoopTST:
        CMP	R7,#24
        BNE     CBMLoop
        ADD	R6,R6,#1
        LDRB	R7,[R11,R6]
        AND	R7,R7,#0x0F
        AND	R8,R0,#0x0F
        ADD	R7,R7,R8
        MOV	R7,R7,LSR#1
        STRB	R7,[R11,R6]
        B       CBMLoop
@----------------------------------

Q8E:                            @ STX AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        STRB    R1,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R1,#1
        BNE     CBMLoopTST
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop

@----------------------------------

Q90:                            @ BCC
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#1
        BNE     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop

@----------------------------------

Q91:                            @ STA(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R6,R7,R8,LSL#8
        ADD     R6,R6,R2
        STRB    R0,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R0,#1
        BNE     CBMLoop
        AND     R7,R6,#0xFF
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop
        
@----------------------------------

Q94:                            @ STY ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        STRB    R2,[R11,R6]
        B       CBMLoop

@----------------------------------

Q95:                            @ STA ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        STRB    R0,[R11,R6]
        B       CBMLoop

@----------------------------------

Q96:                            @ STX ZPY
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R2
        AND     R6,R6,#0xFF
        STRB    R1,[R11,R6]
        B       CBMLoop

@----------------------------------

Q98:                            @ TYA
        ADD     R3,R3,#1
        MOVS    R0,R2
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

Q99:                            @ STA ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        ADD     R6,R6,R2
        STRB    R0,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R0,#1
        BNE     CBMLoop
        AND     R7,R6,#0xFF
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop

@----------------------------------

Q9A:                            @ TXS_(KEINE FLAGBEEINFLUSSUNG!)
        ADD     R3,R3,#1
        MOV     R5,R1
        B       CBMLoop
        
@----------------------------------

Q9D:                            @ STA ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        ADD     R6,R6,R1
        STRB    R0,[R11,R6]
        CMP     R8,#0xD4
        BNE     CBMLoop
        TST     R0,#1
        BNE     CBMLoop
        AND     R7,R6,#0xFF
        CMP     R7,#4
        STREQB  R7,[R10,#0]
        CMP     R7,#4+7
        STREQB  R7,[R10,#1]
        CMP     R7,#4+14
        STREQB  R7,[R10,#2]
        B       CBMLoop
        
@----------------------------------

QA0:                            @ LDY#
        LDRB    R2,[R3,#1]
        ADD     R3,R3,#2
        AND     R4,R4,#b_01111101
        CMP     R2,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QA1:                            @ LDA(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R0,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QA2:                            @ LDX#
        LDRB    R1,[R3,#1]
        ADD     R3,R3,#2
        AND     R4,R4,#b_01111101
        CMP     R1,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QA4:                            @ LDY ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R2,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R2,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QA5:                            @ LDA ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R0,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QA6:                            @ LDX ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R1,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R1,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QA8:                            @ TAY
        ADD     R3,R3,#1
        MOVS    R2,R0
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QA9:                            @ LDA#
        LDRB    R0,[R3,#1]
        ADD     R3,R3,#2
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QAA:                            @ TAX
        ADD     R3,R3,#1
        MOVS    R1,R0
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QAC:                            @ LDY AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R2,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R2,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QAD:                            @ LDA AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R0,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QAE:                            @ LDX AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R1,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R1,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QB0:                            @ BCS
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#1
        BEQ     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop
        
@----------------------------------

QB1:                            @ LDA(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R0,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QB4:                            @ LDY ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R2,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R2,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QB5:                            @ LDA ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R0,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QB6:                            @ LDX ZPY
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R2
        AND     R6,R6,#0xFF
        LDRB    R1,[R11,R6]
        AND     R4,R4,#b_01111101
        CMP     R1,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QB8:                            @ CLV
        ADD     R3,R3,#1
        BIC     R4,R4,#b_01000000
        B       CBMLoop
        
@----------------------------------

QB9:                            @ LDA ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R0,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QBA:                            @ TSX
        ADD     R3,R3,#1
        MOVS    R1,R5
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QBC:                            @ LDY ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R2,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R2,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QBD:                            @ LDA ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R0,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R0,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QBE:                            @ LDX ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R1,[R11,R7]
        AND     R4,R4,#b_01111101
        CMP     R1,#0
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QC0:                            @ CPY#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        
        CMP     R2,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        
        B       CBMLoop

@----------------------------------

QC1:                            @ CMP(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop
        
@----------------------------------

QC4:                            @ CPY ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        
        CMP     R2,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop
        
@----------------------------------

QC5:                            @ CMP ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QC6:                            @ DEC ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        SUB     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QC8:                            @ INY
        ADD     R3,R3,#1
        ADD     R2,R2,#1
        ANDS    R2,R2,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R2,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QC9:                            @ CMP#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QCA:                            @ DEX
        ADD     R3,R3,#1
        SUB     R1,R1,#1
        ANDS    R1,R1,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QCC:                            @ CPY AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        CMP     R2,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QCD:                            @ CMP AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QCE:                            @ DEC AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        LDRB    R7,[R11,R6]
        SUB     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QD0:                            @ BNE
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_00000010
        BNE     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop

@----------------------------------

QD1:                            @ CMP(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QD5:                            @ CMP ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        LDRB    R6,[R11,R6]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QD6:                            @ DEC ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        SUB     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QD8:                            @ CLD
        ADD     R3,R3,#1
        BIC     R4,R4,#b_00001000
        B       CBMLoop

@----------------------------------

QD9:                            @ CMP ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QDD:                            @ CMP ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        
        CMP     R0,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QDE:                            @ DEC ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        ADD     R6,R6,R1
        LDRB    R7,[R11,R6]
        SUB     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QE0:                            @ CPX#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        
        CMP     R1,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QE1:                            @ SBC(,X)
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R6,R6,#1
        AND     R6,R6,#0xFF
        LDRB    R8,[R11,R6]
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QE4:                            @ CPX ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        
        CMP     R1,R6		@ KANN EVTL. R6,R2 SEIN!
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001
        
        B       CBMLoop

@----------------------------------

QE5:                            @ SBC ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R6,[R11,R6]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QE6:                            @ INC ZP
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        LDRB    R7,[R11,R6]
        ADD     R7,R7,#1
        STRB    R7,[R11,R6]
        AND     R4,R4,#b_01111101
        ANDS    R7,R7,#0xFF
        ORREQ   R4,R4,#2
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QE8:                            @ INX
        ADD     R3,R3,#1
        ADD     R1,R1,#1
        ANDS    R1,R1,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R1,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QE9:                            @ SBC#
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

QEA:                            @ NOP
        ADD     R3,R3,#1
        B       CBMLoop

@----------------------------------

QEC:                            @ CPX AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        
        CMP     R1,R6
        AND     R4,R4,#b_01111100
        ORRMI   R4,R4,#b_10000000
        ORREQ   R4,R4,#b_00000010
        ORRCS   R4,R4,#b_00000001

        B       CBMLoop

@----------------------------------

QED:                            @ SBC AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        LDRB    R6,[R11,R7]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QEE:                            @ INC AB
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        LDRB    R7,[R11,R6]
        ADD     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QF0:                            @ BEQ
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        TST     R4,#b_00000010
        BEQ     CBMLoop
        TST     R6,#b_10000000
        ADD     R3,R3,R6
        SUBNE   R3,R3,#0x100
        B       CBMLoop
        
@----------------------------------

QF1:                            @ SBC(),Y
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R11
        LDRB    R7,[R6]
        LDRB    R8,[R6,#1]
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QF5:                            @ SBC ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R6,[R11,R6]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QF6:                            @ INC ZPX
        LDRB    R6,[R3,#1]
        ADD     R3,R3,#2
        ADD     R6,R6,R1
        AND     R6,R6,#0xFF
        LDRB    R7,[R11,R6]
        ADD     R7,R7,#1
        STRB    R7,[R11,R6]
        AND     R4,R4,#b_01111101
        ANDS    R7,R7,#0xFF
        ORREQ   R4,R4,#2
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QF8:                            @ SED
        ADD     R3,R3,#1
        ORR     R4,R4,#b_00001000
        B       CBMLoop

@----------------------------------

QF9:                            @ SBC ABY
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R2
        LDRB    R6,[R11,R7]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QFD:                            @ SBC ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R7,R7,R8,LSL#8
        ADD     R7,R7,R1
        LDRB    R6,[R11,R7]
        SUB     R0,R0,R6
        AND     R8,R4,#1
        EOR     R8,R8,#1
        SUB     R0,R0,R8
        AND     R4,R4,#b_01111100
        CMP     R0,#0
        ORRPL   R4,R4,#1
        ORREQ   R4,R4,#b_00000010
        AND     R0,R0,#0xFF
        CMP     R0,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop

@----------------------------------

QFE:                            @ INC ABX
        LDRB    R7,[R3,#1]
        LDRB    R8,[R3,#2]
        ADD     R3,R3,#3
        ADD     R6,R7,R8,LSL#8
        ADD     R6,R6,R1
        LDRB    R7,[R11,R6]
        ADD     R7,R7,#1
        STRB    R7,[R11,R6]
        ANDS    R7,R7,#0xFF
        AND     R4,R4,#b_01111101
        ORREQ   R4,R4,#b_00000010
        CMP     R7,#0x80
        ORRGE   R4,R4,#b_10000000
        B       CBMLoop
        
@----------------------------------

.END
	